# Copyright 2015 Hewlett-Packard
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import sys

from freezer import __version__ as FREEZER_VERSION
from freezer.utils import winutils
from oslo_config import cfg
from oslo_log import log

CONF = cfg.CONF

if winutils.is_windows():
    DEFAULT_FREEZER_SCHEDULER_CONF_D = r'C:\.freezer\scheduler\conf.d'
else:
    DEFAULT_FREEZER_SCHEDULER_CONF_D = '/etc/freezer/scheduler/conf.d'


def add_filter():

    class NoLogFilter(log.logging.Filter):
        def filter(self, record):
            return False

    log_filter = NoLogFilter()
    log.logging.getLogger("apscheduler.scheduler").\
        addFilter(log_filter)
    log.logging.getLogger("apscheduler.executors.default").\
        addFilter(log_filter)
    log.logging.getLogger("requests.packages.urllib3.connectionpool").\
        addFilter(log_filter)


def env(*args, **kwargs):
    for v in args:
        value = os.environ.get(v, None)
        if value:
            return value
    return kwargs.get('default', '')


def get_common_opts():
    scheduler_conf_d = os.environ.get('FREEZER_SCHEDULER_CONF_D',
                                      DEFAULT_FREEZER_SCHEDULER_CONF_D)

    _COMMON = [
        cfg.StrOpt('client-id',
                   dest='client_id',
                   short='c',
                   help='Specifies the client_id used when contacting the '
                        'service.\n If not specified it will be automatically '
                        'created \n using the project-id and the machine '
                        'hostname.'),
        cfg.BoolOpt('no-api',
                    default=False,
                    dest='no_api',
                    short='n',
                    help='Prevents the scheduler from using the api service'),
        cfg.StrOpt('conf',
                   default=scheduler_conf_d,
                   dest='jobs_dir',
                   short='f',
                   help='Used to store/retrieve files on local storage, '
                        'including those exchanged with the api service. '
                        'Default value is {0} (Env: FREEZER_SCHEDULER_CONF_D)'
                   .format(scheduler_conf_d)),
        cfg.IntOpt('interval',
                   default=60,
                   dest='interval',
                   short='i',
                   help='Specifies the api-polling interval in seconds. '
                        'Defaults to 60 seconds'),
        cfg.BoolOpt('no-daemon',
                    default=False,
                    dest='no_daemon',
                    help='Prevents the scheduler from running in daemon mode'),
        cfg.BoolOpt('insecure',
                    default=False,
                    dest='insecure',
                    help='Initialize freezer scheduler with insecure mode'),
        cfg.BoolOpt('disable-exec',
                    default=False,
                    dest='disable_exec',
                    help='Allow Freezer Scheduler to deny jobs that execute '
                         'commands for security reasons'),
        cfg.IntOpt('concurrent_jobs',
                   default=1,
                   dest='concurrent_jobs',
                   help='Number of jobs that can be executed at the'
                        ' same time'),
        cfg.BoolOpt('enable-v1-api',
                    default=False,
                    dest='enable_v1_api',
                    help='Use freezer-api v1 interface if enable_v1_api'
                         ' is set as True, otherwise use freezer-api'
                         ' v2 interface'),
    ]

    return _COMMON


def build_os_options():
    osclient_opts = [
        cfg.StrOpt('os-username',
                   default=env('OS_USERNAME'),
                   help='Name used for authentication with the OpenStack '
                        'Identity service. Defaults to env[OS_USERNAME].',
                   dest='os_username'),
        cfg.StrOpt('os-password',
                   default=env('OS_PASSWORD'),
                   help='Password used for authentication with the OpenStack '
                        'Identity service. Defaults to env[OS_PASSWORD].',
                   dest='os_password'),
        cfg.StrOpt('os-project-id',
                   default=env('OS_PROJECT_ID'),
                   help='Project id to scope to. Defaults to '
                        'env[OS_PROJECT_ID].',
                   dest='os_project_id'),
        cfg.StrOpt('os-project-name',
                   default=env('OS_PROJECT_NAME'),
                   help='Project name to scope to. Defaults to '
                        'env[OS_PROJECT_NAME].',
                   dest='os_project_name'),
        cfg.StrOpt('os-project-domain-id',
                   default=env('OS_PROJECT_DOMAIN_ID'),
                   help='Domain ID containing project. Defaults to '
                        'env[OS_PROJECT_DOMAIN_ID].',
                   dest='os_project_domain_id'),
        cfg.StrOpt('os-project-domain-name',
                   default=env('OS_PROJECT_DOMAIN_NAME'),
                   help='Domain name containing project. Defaults to '
                        'env[OS_PROJECT_DOMAIN_NAME].',
                   dest='os_project_domain_name'),
        cfg.StrOpt('os-user-domain-id',
                   default=env('OS_USER_DOMAIN_ID'),
                   help='User\'s domain id. Defaults to '
                        'env[OS_USER_DOMAIN_ID].',
                   dest='os_user_domain_id'),
        cfg.StrOpt('os-user-domain-name',
                   default=env('OS_USER_DOMAIN_NAME'),
                   help='User\'s domain name. Defaults to '
                        'env[OS_USER_DOMAIN_NAME].',
                   dest='os_user_domain_name'),
        cfg.StrOpt('os-auth-url',
                   default=env('OS_AUTH_URL'),
                   help='Specify the Identity endpoint to use for '
                        'authentication. Defaults to env[OS_AUTH_URL].',
                   dest='os_auth_url'),
        cfg.StrOpt('os-backup-url',
                   default=env('OS_BACKUP_URL'),
                   help='Specify the Freezer backup service endpoint to use. '
                        'Defaults to env[OS_BACKUP_URL].',
                   dest='os_backup_url'),
        cfg.StrOpt('os-region-name',
                   default=env('OS_REGION_NAME'),
                   help='Specify the region to use. Defaults to '
                        'env[OS_REGION_NAME].',
                   dest='os_region_name'),
        cfg.StrOpt('os-token',
                   default=env('OS_TOKEN'),
                   help='Specify an existing token to use instead of '
                        'retrieving one via authentication '
                        '(e.g. with username & password). Defaults '
                        'to env[OS_TOKEN].',
                   dest='os_token'),
        cfg.StrOpt('os-endpoint-type',
                   choices=['public', 'publicURL', 'internal', 'internalURL',
                            'admin', 'adminURL'],
                   default=env('OS_ENDPOINT_TYPE') or 'public',
                   help='Endpoint type to select. Valid endpoint types: '
                        '"public" or "publicURL", "internal" or "internalURL",'
                        ' "admin" or "adminURL". Defaults to '
                        'env[OS_ENDPOINT_TYPE] or "public"',
                   dest='os_endpoint_type'),
        cfg.StrOpt('os-cert',
                   default=env('OS_CERT'),
                   help='Specify a cert file to use in verifying a TLS '
                        '(https) server certificate',
                   dest='os_cert'),
        cfg.StrOpt('os-cacert',
                   default=env('OS_CACERT'),
                   help='Specify a CA bundle file to use in verifying a TLS '
                        '(https) server certificate. Defaults to',
                   dest='os_cacert'),
    ]

    return osclient_opts


def parse_args(choices):
    default_conf = cfg.find_config_files('freezer', 'scheduler', '.conf')
    CONF.register_cli_opts(get_common_opts())
    CONF.register_cli_opts(build_os_options())
    log.register_options(CONF)

    positional = [
        cfg.StrOpt('action',
                   choices=choices,
                   help='{0}'.format(choices), positional=True),

    ]
    CONF.register_cli_opts(positional)
    CONF(args=sys.argv[1:],
         project='freezer-scheduler',
         default_config_files=default_conf,
         version=FREEZER_VERSION
         )


def setup_logging():
    _DEFAULT_LOG_LEVELS = ['amqp=WARN', 'amqplib=WARN', 'boto=WARN',
                           'stevedore=WARN', 'oslo_log=INFO',
                           'iso8601=WARN', 'urllib3.connectionpool=WARN',
                           'requests.packages.urllib3.connectionpool=WARN',
                           'websocket=WARN', 'keystonemiddleware=WARN']

    _DEFAULT_LOGGING_CONTEXT_FORMAT = ('%(asctime)s.%(msecs)03d %(process)d '
                                       '%(levelname)s %(name)s '
                                       '[%(request_id)s %(user_identity)s] '
                                       '%(instance)s%(message)s')
    add_filter()
    log.set_defaults(_DEFAULT_LOGGING_CONTEXT_FORMAT, _DEFAULT_LOG_LEVELS)
    log.setup(CONF, 'freezer-scheduler', version=FREEZER_VERSION)


def list_opts():
    _opt = {
        None: get_common_opts()
    }
    return _opt.items()



"""
(c) Copyright 2018 ZTE Corporation.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

import ftplib
import os
import shutil
import socket
import tempfile

from freezer.storage import fslike
from freezer.utils import utils
from oslo_log import log
from oslo_serialization import jsonutils as json

CHUNK_SIZE = 32768
LOG = log.getLogger(__name__)


class BaseFtpStorage(fslike.FsLikeStorage):
    """
    :type ftp: ftplib
    """
    _type = 'ftpbase'

    def __init__(self, storage_path, remote_pwd,
                 remote_username, remote_ip, port, max_segment_size):
        """
            :param storage_path: directory of storage
            :type storage_path: str
            :return:
            """
        self.remote_username = remote_username
        self.remote_pwd = remote_pwd
        self.remote_ip = remote_ip
        self.port = port
        self.ftp = None
        self._validate()
        self.init()
        super(BaseFtpStorage, self).__init__(
            storage_path=storage_path,
            max_segment_size=max_segment_size)

    def _validate(self):
        """
        Validates if all parameters required to ssh are available.
        :return: True or raises ValueError
        """
        if not self.remote_ip:
            raise ValueError('Please provide --ftp-host value.')
        elif not self.remote_username:
            raise ValueError('Please provide --ftp-username value.')
        elif not self.remote_pwd:
            raise ValueError('Please provide remote password.'
                             '--ftp-password argument.')
        return True

    def init(self):
        pass

    def _create_tempdir(self):
        try:
            tmpdir = tempfile.mkdtemp()
        except Exception:
            LOG.error("Unable to create a tmp directory")
            raise
        return tmpdir

    def rmtree(self, path):
        LOG.info("ftp rmtree path=%s" % path)
        files = []
        self.ftp.dir(path, files.append)
        LOG.info('rm files=%s' % files)
        for f in files:
            attr = f.split()[0]
            file_name = f.split()[-1]
            filepath = utils.path_join(path, file_name)
            if attr.startswith('d'):
                self.rmtree(filepath)
            else:
                self.ftp.delete(filepath)
        self.ftp.rmd(path)

    def create_dirs(self, path):
        """Change to this directory, recursively making new folders if needed.
        Returns True if any folders were created."""
        LOG.info("ftp create_dirs path=%s" % path)
        if path == '/':
            # absolute path so change directory to root
            self.ftp.cwd('/')
            return
        if path == '':
            # top-level relative directory must exist
            return
        try:
            self.ftp.cwd(path)  # sub-directory exists
        except ftplib.all_errors as e:
            LOG.info("ftp create dirs failed %s" % e)
            dirname, basename = os.path.split(path.rstrip('/'))
            LOG.info("ftp create_dirs dirname=%s basename=%s"
                     % (dirname, basename))
            self.create_dirs(dirname)  # make parent directories
            self.ftp.mkd(basename)  # sub-directory missing, so created it
            self.ftp.cwd(basename)
            return True

    def write_backup(self, rich_queue, backup):
        """
        Stores backup in storage
        :type rich_queue: freezer.streaming.RichQueue
        :type backup: freezer.storage.base.Backup
        """
        try:
            tmpdir = tempfile.mkdtemp()
        except Exception:
            LOG.error("Unable to create a tmp directory")
            raise

        try:
            data_meta = utils.path_join(tmpdir, "data_meta")
            LOG.info("ftp write data_meta %s" % data_meta)
            backup = backup.copy(storage=self)
            path = backup.data_path
            self.create_dirs(path.rsplit('/', 1)[0])

            with open(data_meta, mode='wb') as b_file:
                for message in rich_queue.get_messages():
                    b_file.write(message)

            self.put_file(data_meta, path)
        finally:
            shutil.rmtree(tmpdir)

    def get_file(self, from_path, to_path):
        LOG.info("ftp get_file from_path=%s to_path=%s" % (from_path, to_path))
        try:
            dir = self.ftp.pwd()
            LOG.info("ftp get file dir %s" % dir)
        except (ftplib.all_errors, socket.error) as e:
            LOG.info("ftp get file failed %s try again" % e)
            self.init()
        try:
            file = open(to_path, 'wb')
            msg = self.ftp.retrbinary('RETR ' + from_path,
                                      file.write, 8192)
            # 226
            LOG.info("FTP GET %s, ret=%s" % (from_path, msg))
        except ftplib.all_errors as e:
            file.close()
            self.ftp.quit()
            LOG.info("ftp get file error %s" % e)
            raise e
        file.close()

    def put_file(self, from_path, to_path):
        LOG.info("ftp put_file from_path=%s to_path=%s" % (from_path, to_path))
        try:
            dir = self.ftp.pwd()
            LOG.info("ftp put file dir %s" % dir)
        except (ftplib.all_errors, socket.error) as e:
            LOG.info("ftp put file failed %s try again" % e)
            self.init()
        try:
            file = open(from_path, 'rb')
            msg = self.ftp.storbinary('STOR ' + to_path,
                                      file, 8192)
            # 226
            LOG.info("FTP PUT %s, ret=%s" % (from_path, msg))
        except ftplib.all_errors as e:
            self.ftp.quit()
            LOG.info("ftp put file error %s" % e)
            raise e
        file.close()

    def listdir(self, directory):
        LOG.info("ftp listdir directory=%s" % directory)
        try:
            # paramiko SFTPClient.listdir_attr returns
            # directories in arbitarary order, so we should
            # sort results of this command
            ret = self.ftp.cwd(directory)
            LOG.info('ftp listdir cwd ret=%s' % ret)
            res = self.ftp.nlst()
            LOG.info('ftp listdir res=%s' % res)
            return sorted(res)
        except ftplib.error_perm as e:
            LOG.info("ftp listdir error %s" % e)
            if '550' in e[0]:
                return list()
            else:
                raise

    def open(self, path, mode):
        pass

    def backup_blocks(self, backup):
        LOG.info("ftp backup_blocks ")
        self.init()
        # should recreate ssh for new process
        tmpdir = self._create_tempdir()
        try:
            data = utils.path_join(tmpdir, "data")
            LOG.info("backup_blocksa datadown=%s" % data)
            self.get_file(backup.data_path, data)
            with open(data, 'rb') as backup_file:
                while True:
                    chunk = backup_file.read(self.max_segment_size)
                    if chunk == '':
                        break
                    if len(chunk):
                        yield chunk
        finally:
            shutil.rmtree(tmpdir)

    def add_stream(self, stream, package_name, headers=None):
        """
        :param stream: data
        :param package_name: path
        :param headers: backup metadata information
        :return:
        """
        tmpdir = self._create_tempdir()
        LOG.info('add stream')
        try:
            split = package_name.rsplit('/', 1)
            # create backup_basedir
            backup_basedir = "{0}/{1}".format(self.storage_path,
                                              package_name)
            self.create_dirs(backup_basedir)
            # define backup_data_name
            backup_basepath = "{0}/{1}".format(backup_basedir,
                                               split[0])
            backup_metadata = "%s/metadata" % backup_basedir
            # write backup to backup_basepath
            data_backup = utils.path_join(tmpdir, "data_backup")
            with open(data_backup, 'wb') as backup_file:
                for el in stream:
                    backup_file.write(el)
            self.put_file(data_backup, backup_basepath)
            # write data matadata to backup_metadata
            metadata = utils.path_join(tmpdir, "metadata")
            with open(metadata, 'wb') as backup_meta:
                backup_meta.write(json.dumps(headers))
            self.put_file(metadata, backup_metadata)
        finally:
            shutil.rmtree(tmpdir)


class FtpStorage(BaseFtpStorage):
    """
    :type ftp: ftplib.FTP()
    """
    _type = 'ftp'

    def __init__(self, storage_path, remote_pwd,
                 remote_username, remote_ip, port, max_segment_size):
        """
            :param storage_path: directory of storage
            :type storage_path: str
            :return:
            """
        super(FtpStorage, self).__init__(storage_path,
                                         remote_pwd,
                                         remote_username,
                                         remote_ip,
                                         port,
                                         max_segment_size)

    def init(self):
        try:
            ftp = ftplib.FTP()
            ftp.set_pasv(True)
            # socket.setdefaulttimeout(60)
            ftp.connect(self.remote_ip, self.port, 60)
            ftp.login(self.remote_username, self.remote_pwd)
            nfiles = ftp.nlst()
            LOG.info("ftp nlst result=%s" % nfiles)
        except socket.error as e:
            LOG.info("ftp socket error=%s" % e)
            ftp.set_pasv(False)
        except ftplib.all_errors as e:  # socket.error
            msg = "create ftp failed error=%s" % e
            LOG.info(msg)
            raise Exception(msg)
        # we should keep link to ssh to prevent garbage collection
        self.ftp = ftp


class FtpsStorage(BaseFtpStorage):
    """
    :type ftps: ftplib.FTP_TLS()
    """
    _type = 'ftps'

    def __init__(self, storage_path, remote_pwd,
                 remote_username, remote_ip, port, max_segment_size,
                 keyfile, certfile):
        """
            :param storage_path: directory of storage
            :type storage_path: str
            :return:
            """
        self.keyfile = keyfile
        self.certfile = certfile
        LOG.info("key=%s cer=%s" % (self.keyfile, self.certfile))
        super(FtpsStorage, self).__init__(storage_path, remote_pwd,
                                          remote_username, remote_ip,
                                          port, max_segment_size)

    def init(self):
        try:
            ftps = ftplib.FTP_TLS(keyfile=self.keyfile,
                                  certfile=self.certfile)
            ftps.set_pasv(True)
            ftps.connect(self.remote_ip, self.port, 60)
            ftps.login(self.remote_username, self.remote_pwd)
            msg = ftps.prot_p()
            LOG.info("ftps encrypt %s, ret=%s" % (self.remote_ip, msg))
            nfiles = ftps.nlst()
            LOG.info("ftps nlst result=%s" % nfiles)
        except socket.error as e:
            LOG.info("ftps socket error=%s" % e)
            ftps.set_pasv(False)
        except ftplib.all_errors as e:
            msg = "create ftps failed error=%s" % e
            LOG.info(msg)
            raise Exception(msg)
        self.ftp = ftps



# (c) Copyright 2014,2015 Hewlett-Packard Development Company, L.P.
# (c) Copyright 2016 Hewlett-Packard Enterprise Development Company, L.P.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Freezer general utils functions
"""

import threading

from oslo_log import log
import queue


LOG = log.getLogger(__name__)


class Wait(Exception):
    pass


class RichQueue(object):
    """
        :type data_queue: Queue.Queue
    """
    def __init__(self, size=2):
        """
        :type size: int
        :return:
        """
        self.data_queue = queue.Queue(maxsize=size)
        # transmission changes in atomic way so no synchronization needed
        self.finish_transmission = False
        self.is_force_stop = False

    def finish(self):
        self.finish_transmission = True

    def force_stop(self):
        self.is_force_stop = True

    def empty(self):
        return self.data_queue.empty()

    def get(self):
        try:
            res = self.data_queue.get(timeout=1)
            self.data_queue.task_done()
            return res
        except queue.Empty:
            raise Wait()

    def check_stop(self):
        if self.is_force_stop:
            raise Exception("Forced stop")

    def put_messages(self, messages):
        for message in messages:
            self.put(message)
        self.finish()

    def has_more(self):
        self.check_stop()
        return not self.finish_transmission or not self.data_queue.empty()

    def put(self, message):
        while True:
            try:
                self.data_queue.put(message, timeout=1)
                break
            except queue.Full:
                self.check_stop()

    def get_messages(self):
        while self.has_more():
            try:
                yield self.get()
            except Wait:
                self.check_stop()


class QueuedThread(threading.Thread):
    def __init__(self, target, rich_queue, exception_queue,
                 args=(), kwargs=None):
        """
            :type args: collections.Iterable
            :type kwargs: dict
            :type target: () -> ()
            :type rich_queue: RichQueue
            """
        self.args = args
        kwargs = kwargs or {}
        self.rich_queue = rich_queue
        self._exception_queue = exception_queue
        kwargs["rich_queue"] = rich_queue
        super(QueuedThread, self).__init__(target=target, args=args,
                                           kwargs=kwargs)

    def run(self):
        try:
            super(QueuedThread, self).run()
        except Exception as e:
            LOG.exception(e)
            self._exception_queue.put_nowait(e)
            self.rich_queue.force_stop()
            # Thread will exit at this point.
            # @todo print the error using traceback.print_exc(file=sys.stdout)
            raise



# (c) Copyright 2014,2015 Hewlett-Packard Development Company, L.P.
# (c) Copyright 2016 Hewlett-Packard Enterprise Development Company, L.P
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Freezer restore.py related tests
"""
from unittest import mock

from freezer.openstack import restore
from freezer.tests import commons


class Image(object):
    def __init__(self):
        self.id = 'test'


class TestRestore(commons.FreezerBaseTestCase):
    def setUp(self):
        super(TestRestore, self).setUp()

    def test_restore_cinder_by_glance(self):
        backup_opt = commons.BackupOpt1()
        restore.RestoreOs(backup_opt.client_manager, backup_opt.container,
                          backup_opt.storage)

    def test_restore_cinder_by_glance_from_local(self):
        backup_opt = commons.BackupOpt1()
        restore.RestoreOs(backup_opt.client_manager, backup_opt.container,
                          'local')

    def test_restore_cinder_with_backup_id(self):
        backup_opt = commons.BackupOpt1()
        ros = restore.RestoreOs(backup_opt.client_manager,
                                backup_opt.container, backup_opt.storage)
        ros.restore_cinder(35, 34, 33)

    def test_restore_cinder_without_backup_id(self):
        backup_opt = commons.BackupOpt1()
        ros = restore.RestoreOs(backup_opt.client_manager,
                                backup_opt.container, backup_opt.storage)
        ros.restore_cinder(35, 34)

    def test_restore_nova(self):
        backup_opt = commons.BackupOpt1()
        restore.RestoreOs(backup_opt.client_manager, backup_opt.container,
                          backup_opt.storage)

    def test_restore_nova_from_local(self):
        backup_opt = commons.BackupOpt1()
        restore.RestoreOs(backup_opt.client_manager, backup_opt.container,
                          'local')

    def test_restore_cinder(self):
        storage = mock.MagicMock()
        storage.type = 'swift'
        backup1 = mock.MagicMock()
        backup1.created_at = "2020-08-31T16:32:30"
        backup2 = mock.MagicMock()
        backup2.created_at = "2020-08-31T16:32:31"
        cinder_client = mock.MagicMock()
        cinder_client.backups.list.return_value = [backup1, backup2]
        cinder_client.restores.restore.return_value = 'test'
        client_manager = mock.MagicMock()
        client_manager.get_cinder.return_value = cinder_client
        restore_os = restore.RestoreOs(client_manager, '/root/test/', storage)
        result = restore_os.restore_cinder(restore_from_timestamp=1598862750)
        self.assertIsNone(result)
        result = restore_os.restore_cinder(restore_from_timestamp=1598862749)
        self.assertIsNone(result)

    @mock.patch('shutil.rmtree')
    @mock.patch('tempfile.mkdtemp')
    @mock.patch('oslo_serialization.jsonutils.load')
    @mock.patch('os.listdir')
    @mock.patch('builtins.open')
    @mock.patch('freezer.utils.utils.ReSizeStream')
    @mock.patch('freezer.utils.utils.S3ResponseStream')
    def test_create_image(self, mock_s3stream, mock_resize, mock_open,
                          mock_list, mock_load, mock_mkdtemp, mock_rmtree):
        image = Image()
        storage = mock.MagicMock()
        storage.type = 's3'
        storage.get_object.return_value = {'Body': '{"test": "test_info"}',
                                           'ContentLength': 3}
        storage.get_object_prefix.return_value = 'test'
        storage.list_all_objects.return_value = [{'Key': '/12345'},
                                                 {'Key': '/12346'}]
        client_manager = mock.MagicMock()
        client_manager.create_image.return_value = image
        restore_os = restore.RestoreOs(client_manager, '/root/test/', storage)
        result1, result2 = restore_os._create_image('/root', 12344)
        self.assertEqual(result1, {"test": "test_info"})
        self.assertEqual(result2, image)

        storage.get_object_prefix.return_value = ''
        result1, result2 = restore_os._create_image('/root', 12344)
        self.assertEqual(result1, {"test": "test_info"})
        self.assertEqual(result2, image)

        mock_open.side_effect = Exception("error")
        storage.type = 'local'
        mock_list.return_value = ['12345']
        restore_os = restore.RestoreOs(client_manager, '/root/test', storage)
        try:
            restore_os._create_image('/root', 12344)
        except BaseException as e:
            self.assertEqual(str(e), "Failed to open image file"
                                     " /root/test//root/12345//root")

        mock_open.side_effect = 'test'
        mock_load.return_value = 'test'
        result1, result2 = restore_os._create_image('/root', 12344)
        self.assertEqual(result1, 'test')
        self.assertEqual(result2, image)

        storage.type = 'ssh'
        storage.open.side_effect = Exception("error")
        storage.listdir.return_value = ['12345']
        restore_os = restore.RestoreOs(client_manager, '/root/test', storage)
        try:
            restore_os._create_image('/root', 12344)
        except BaseException as e:
            self.assertEqual(str(e), "Failed to open remote image file "
                                     "/root/test//root/12345//root")

        storage.open.side_effect = 'test'
        storage.read_metadata_file.return_value = '{"test": "test_info"}'
        restore_os = restore.RestoreOs(client_manager, '/root/test', storage)
        result1, result2 = restore_os._create_image('/root', 12344)
        self.assertEqual(result1, {"test": "test_info"})
        self.assertEqual(result2, image)

        storage.type = 'ftp'
        storage.listdir.return_value = ['12345']
        mock_mkdtemp.side_effect = Exception('error')
        restore_os = restore.RestoreOs(client_manager, '/root/test', storage)
        try:
            restore_os._create_image('/root', 12344)
        except Exception as e:
            self.assertEqual(str(e), "Unable to create a tmp directory")

        mock_mkdtemp.side_effect = "success"
        mock_rmtree.return_value = "success"
        mock_open.side_effect = 'test'
        result1, result2 = restore_os._create_image('/root', 12344)
        self.assertEqual(result1, 'test')
        self.assertEqual(result2, image)

    def test_get_backups_exception(self):
        storage = mock.MagicMock()
        storage.type = 'ss3'
        client_manager = mock.MagicMock()
        restore_os = restore.RestoreOs(client_manager, '/root/test/', storage)
        try:
            restore_os._get_backups('/root', 12347)
        except BaseException as e:
            self.assertEqual(str(e), "ss3 storage type is not supported at the"
                                     " moment. Try local, SWIFT, SSH(SFTP),"
                                     " FTP or FTPS ")

        storage = mock.MagicMock()
        storage.type = 's3'
        storage.list_all_objects.return_value = [{'Key': '/12345'},
                                                 {'Key': '/12346'}]
        restore_os = restore.RestoreOs(client_manager, '/root/test/', storage)
        try:
            restore_os._get_backups('/root', 12347)
        except BaseException as e:
            self.assertEqual(str(e), "Cannot find backups for"
                                     " path: root/test///root")



# (c) Copyright 2014-2016 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Freezer LVM related functions
"""

import os
import re
import subprocess

from oslo_log import log
from oslo_utils import uuidutils

from freezer.common import config as freezer_config
from freezer.utils import utils

LOG = log.getLogger(__name__)


def lvm_snap_remove(backup_opt_dict):
    """
    Unmount the snapshot and removes it

    :param backup_opt_dict.lvm_dirmount: mount point of the snapshot
    :param backup_opt_dict.lvm_volgroup: volume group to which the lv belongs
    :param backup_opt_dict.lvm_snapname: name of the snapshot lv
    :return: None, raises on error
    """
    try:
        _umount(backup_opt_dict.lvm_dirmount)
    except Exception as e:
        LOG.warning("Snapshot unmount errror: {0}".format(e))
    lv = os.path.join('/dev',
                      backup_opt_dict.lvm_volgroup,
                      backup_opt_dict.lvm_snapname)
    _lvremove(lv)
    LOG.info('Snapshot volume {0} removed'.format(lv))


def get_vol_fs_type(vol_name):
    """
    The argument need to be a full path lvm name i.e. /dev/vg0/var
    or a disk partition like /dev/sda1. The returnet value is the
    file system type
    """
    if os.path.exists(vol_name) is False:
        err = 'Provided volume name not found: {0} '.format(vol_name)
        LOG.exception(err)
        raise Exception(err)

    file_cmd = '{0} -0 -bLs --no-pad --no-buffer --preserve-date \
    {1}'.format(utils.find_executable("file"), vol_name)
    file_process = subprocess.Popen(
        file_cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
        stderr=subprocess.PIPE, shell=True,
        executable=utils.find_executable("bash"))
    (file_out, file_err) = file_process.communicate()
    file_match = re.search(r'(\S+?) filesystem data', file_out, re.I)
    if file_match is None:
        err = 'File system type not guessable: {0}'.format(file_err)
        LOG.exception(err)
        raise Exception(err)
    else:
        filesys_type = file_match.group(1)
        LOG.info('File system {0} found for volume {1}'.format(
            filesys_type, vol_name))
        return filesys_type.lower().strip()


def lvm_snap(backup_opt_dict):
    """
    Checks the provided parameters and create the lvm snapshot if requested

    The path_to_backup might be adjusted in case the user requested
    a lvm snapshot without specifying an exact path for the snapshot).
    The assumption in this case is that the user wants to use the lvm snapshot
    capability to backup the specified filesystem path, leaving out all
    the rest of the parameters which will guessed and set by freezer.

    :param backup_opt_dict: the configuration dict
    :return: True if the snapshot has been taken, False otherwise
    """
    lvm_uuid = uuidutils.generate_uuid(dashed=False)

    if not backup_opt_dict.lvm_snapname:
        backup_opt_dict.lvm_snapname = \
            "{0}_{1}".format(freezer_config.DEFAULT_LVM_SNAP_BASENAME,
                             lvm_uuid)

    # adjust/check lvm parameters according to provided path_to_backup
    lvm_info = get_lvm_info(backup_opt_dict.path_to_backup)

    if not backup_opt_dict.lvm_volgroup:
        backup_opt_dict.lvm_volgroup = lvm_info['volgroup']

    if not backup_opt_dict.lvm_srcvol:
        backup_opt_dict.lvm_srcvol = lvm_info['srcvol']

    if not backup_opt_dict.lvm_dirmount:
        utils.create_dir(freezer_config.DEFAULT_LVM_MOUNT_BASEDIR)
        backup_opt_dict.lvm_dirmount = \
            "{0}/mount_{1}".format(freezer_config.DEFAULT_LVM_MOUNT_BASEDIR,
                                   lvm_uuid)

    if not validate_lvm_params(backup_opt_dict):
        LOG.info('No LVM requested/configured')
        return False

    utils.create_dir(backup_opt_dict.lvm_dirmount)

    if '%' in backup_opt_dict.lvm_snapsize:
        lvm_size_option = "--extents"
    else:
        lvm_size_option = "--size"

    lvm_create_command = (
        '{0} {1} {2} --snapshot --permission {3} '
        '--name {4} {5}'.format(
            utils.find_executable('lvcreate'),
            lvm_size_option,
            backup_opt_dict.lvm_snapsize,
            ('r' if backup_opt_dict.lvm_snapperm == 'ro'
             else backup_opt_dict.lvm_snapperm),
            backup_opt_dict.lvm_snapname,
            backup_opt_dict.lvm_srcvol))

    lvm_process = subprocess.Popen(
        lvm_create_command, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
        stderr=subprocess.PIPE, shell=True,
        executable=utils.find_executable('bash'))
    (lvm_out, lvm_err) = lvm_process.communicate()

    if lvm_process.returncode:
        raise Exception('lvm snapshot creation error: {0}'.format(lvm_err))

    LOG.debug('{0}'.format(lvm_out))
    LOG.warning('Logical volume "{0}" created'.
                format(backup_opt_dict.lvm_snapname))

    # Guess the file system of the provided source volume and st mount
    # options accordingly
    filesys_type = get_vol_fs_type(backup_opt_dict.lvm_srcvol)
    mount_options = '-o {}'.format(backup_opt_dict.lvm_snapperm)
    if 'xfs' == filesys_type:
        mount_options = ' -onouuid '
    # Mount the newly created snapshot to dir_mount
    abs_snap_name = '/dev/{0}/{1}'.format(
        backup_opt_dict.lvm_volgroup,
        backup_opt_dict.lvm_snapname)
    mount_command = '{0} {1} {2} {3}'.format(
        utils.find_executable('mount'),
        mount_options,
        abs_snap_name,
        backup_opt_dict.lvm_dirmount)
    mount_process = subprocess.Popen(
        mount_command, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
        stderr=subprocess.PIPE, shell=True,
        executable=utils.find_executable('bash'))
    mount_err = mount_process.communicate()[1]
    if 'already mounted' in mount_err:
        LOG.warning('Volume {0} already mounted on {1}\
        '.format(abs_snap_name, backup_opt_dict.lvm_dirmount))
        return True
    if mount_err:
        LOG.error("Snapshot mount error. Removing snapshot")
        lvm_snap_remove(backup_opt_dict)
        raise Exception('lvm snapshot mounting error: {0}'.format(mount_err))
    else:
        LOG.warning(
            'Volume {0} successfully mounted on {1}'.format(
                abs_snap_name, backup_opt_dict.lvm_dirmount))

    # After snapshot is mounted, adjust path_to_backup according the mount
    # point and relative path of the snapshot volume.
    lvm_info = get_lvm_info(backup_opt_dict.path_to_backup)
    backup_opt_dict.path_to_backup = os.path.join(backup_opt_dict.lvm_dirmount,
                                                  lvm_info['snap_path'])

    return True


def get_lvm_info(path):
    """
    Take a file system path as argument as backup_opt_dict.path_to_backup
    and return a list containing lvm_srcvol, lvm_volgroup
    where the path is mounted on.

    :param path: the original file system path where backup needs
        to be executed
    :returns: a dict containing the keys 'volgroup', 'srcvol' and 'snap_path'
    """

    mount_point_path, snap_path = utils.get_mount_from_path(path)

    with open('/proc/mounts', 'r') as mount_fd:
        mount_points = mount_fd.readlines()
        lvm_volgroup, lvm_srcvol, lvm_device = lvm_guess(
            mount_point_path, mount_points, '/proc/mounts')

    if not lvm_device:
        mount_process = subprocess.Popen(
            ['mount'], stdout=subprocess.PIPE, stderr=subprocess.PIPE,
            env=os.environ)
        mount_out, mount_err = mount_process.communicate()
        mount_points = [m for m in mount_out.split('\n') if m]
        lvm_volgroup, lvm_srcvol, lvm_device = lvm_guess(
            mount_point_path, mount_points, 'mount')

    if not lvm_device:
        raise Exception(
            'Cannot find {0} in {1}, please provide volume group and '
            'volume name explicitly'.format(mount_point_path, mount_points))

    lvm_params = {'volgroup': lvm_volgroup,
                  'srcvol': lvm_device,
                  'snap_path': snap_path}

    return lvm_params


def lvm_guess(mount_point_path, mount_points, source='/proc/mounts'):
    """Guess lvm vol group and vol name from mount point

    Extract the vol group and vol name from given list
    of mount_points and mount_point_path

    :param mount_point_path: mount path
    :param mount_points: list of currently mounted devices
    :return: a list containing volume group, volume name and full device path
    """

    lvm_volgroup = lvm_srcvol = lvm_device = None
    for mount_line in mount_points:
        if source == '/proc/mounts':
            device, mount_path = mount_line.split(' ')[0:2]
        elif source == 'mount':
            mount_list = mount_line.split(' ')[0:3]
            device = mount_list[0]
            mount_path = mount_list[2]
        if mount_point_path.strip() == mount_path.strip():
            mount_match = re.search(
                r'/dev/mapper/(\w.+?\w)-(\w.+?\w)$', device)
            if mount_match:
                lvm_volgroup = mount_match.group(1).replace('--', '-')
                lvm_srcvol = mount_match.group(2).replace('--', '-')
                lvm_device = u'/dev/{0}/{1}'.format(lvm_volgroup, lvm_srcvol)
                break

    return lvm_volgroup, lvm_srcvol, lvm_device


def validate_lvm_params(backup_opt_dict):
    """
    Validates the parameters and raises in case of missing values

    :param backup_opt_dict:
    :return: False is snapshot is not requested,
             True snapshot is requested and parameters are valid
    """
    if backup_opt_dict.lvm_snapperm not in ('ro', 'rw'):
        raise ValueError('Error: Invalid value for option lvm-snap-perm: '
                         '{}'.format(backup_opt_dict.lvm_snapperm))

    if not backup_opt_dict.path_to_backup:
        raise ValueError('Error: no path-to-backup and '
                         'no lvm-auto-snap provided')

    if not backup_opt_dict.lvm_srcvol and not backup_opt_dict.lvm_volgroup:
        # no lvm parameters provided, assume lvm snapshot is not requested
        return False

    if not backup_opt_dict.lvm_srcvol:
        raise ValueError('Error: no lvm-srcvol and '
                         'no lvm-auto-snap provided')
    if not backup_opt_dict.lvm_volgroup:
        raise ValueError('Error: no lvm-volgroup and '
                         'no lvm-auto-snap provided')

    LOG.info('Source LVM Volume: {0}'.format(
        backup_opt_dict.lvm_srcvol))
    LOG.info('LVM Volume Group: {0}'.format(
        backup_opt_dict.lvm_volgroup))
    LOG.info('Snapshot name: {0}'.format(
        backup_opt_dict.lvm_snapname))
    LOG.info('Snapshot size: {0}'.format(
        backup_opt_dict.lvm_snapsize))
    LOG.info('Directory where the lvm snaphost will be mounted on:'
             ' {0}'.format(backup_opt_dict.lvm_dirmount.strip()))
    LOG.info('Path to backup (including snapshot): {0}'
             .format(backup_opt_dict.path_to_backup))

    return True


def _umount(path):
    # Change dir if we are within the mount point to be removed.
    if os.getcwd().startswith(os.path.normpath(path)):
        os.chdir('/')
    umount_proc = subprocess.Popen('{0} -l -f {1}'.format(
        utils.find_executable('umount'), path),
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE, stderr=subprocess.PIPE,
        shell=True, executable=utils.find_executable('bash'))
    (umount_out, mount_err) = umount_proc.communicate()

    if umount_proc.returncode:
        raise Exception('Impossible to umount {0}. {1}'
                        .format(path, mount_err))
    os.rmdir(path)
    LOG.info('Volume {0} unmounted'.format(path))


def _lvremove(lv):
    for attempt in range(5):
        lvremove_proc = subprocess.Popen(
            '{0} -f {1}'.format(utils.find_executable('lvremove'), lv),
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE, shell=True,
            executable=utils.find_executable('bash'))
        output, error = lvremove_proc.communicate()
        if lvremove_proc.returncode:
            if "contains a filesystem in use" in error:
                LOG.warning("Couldn't remove volume {0}. "
                            "It is still in use.".format(lv))
                log_volume_holding_process(lv)
            else:
                break
        else:
            return
    # Raise if five attempts made or different error than fs in use
    raise Exception('Unable to remove snapshot {0}. {1}'.format(lv, error))


def log_volume_holding_process(lv):
    try:
        # Let's try to provide more information on the failure
        devices = [i.split("\t") for i in subprocess.check_output([
            utils.find_executable('dmsetup'), "ls"]).splitlines()]
        dev_id = [i[1].strip("()").split(":") for i in devices if
                  lv.split("/").pop() in i[0] and
                  not i[0].endswith("cow")][0]
        command = "{} | grep {},{}".format(
            # lsof is quite long, so no need to add a sleep here
            utils.find_executable('lsof'), dev_id[0], dev_id[1])
        process = subprocess.check_output([command], shell=True)
        LOG.warning("Process holding the volume is '{}'".format(process))
    except Exception as e:
        LOG.warning("Could not get informations on the process holding the"
                    " volume: {}".format(str(e)))



# (c) Copyright 2014,2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Freezer main.py related tests
"""

from unittest import mock
from unittest.mock import patch

from freezer import job as jobs
from freezer.tests import commons


class TestJob(commons.FreezerBaseTestCase):
    def setUp(self):
        super(TestJob, self).setUp()

    def test_execute(self):
        opt = commons.BackupOpt1()
        job = jobs.InfoJob(opt, opt.storage)
        assert job.execute() is not None


class TestInfoJob(TestJob):

    def test_execute_nothing_to_do(self):
        backup_opt = commons.BackupOpt1()
        job = jobs.InfoJob(backup_opt, backup_opt.storage)
        job.execute()

    def test_execute_list_containers(self):
        backup_opt = commons.BackupOpt1()
        job = jobs.InfoJob(backup_opt, backup_opt.storage)
        job.execute()


class TestBackupJob(TestJob):
    def setUp(self):
        super(TestBackupJob, self).setUp()

    def test_execute_backup_fs_no_incremental_and_backup_level_raise(self):
        backup_opt = commons.BackupOpt1()
        backup_opt.mode = 'default'
        backup_opt.no_incremental = True
        backup_opt.max_level = None
        backup_opt.always_level = None
        job = jobs.BackupJob(backup_opt, backup_opt.storage)
        self.assertRaises(Exception, job.execute)  # noqa


class TestAdminJob(TestJob):
    def setUp(self):
        super(TestAdminJob, self).setUp()

    def test_execute(self):
        backup_opt = commons.BackupOpt1()
        jobs.AdminJob(backup_opt, backup_opt.storage).execute()


class TestExecJob(TestJob):
    def setUp(self):
        super(TestExecJob, self).setUp()
        # init mock_popen
        self.popen = patch('freezer.utils.exec_cmd.subprocess.Popen')
        self.mock_popen = self.popen.start()
        self.mock_popen.return_value = mock.Mock()
        self.mock_popen.return_value.communicate = mock.Mock()
        self.mock_popen.return_value.communicate.return_value = ['some stderr']

    def tearDown(self):
        super(TestExecJob, self).tearDown()
        self.popen.stop()

    def test_execute_nothing_to_do(self):
        self.mock_popen.return_value.returncode = 0
        backup_opt = commons.BackupOpt1()
        backup_opt.command = 'ls '
        jobs.ExecJob(backup_opt, backup_opt.storage).execute()

    def test_execute_script(self):
        self.mock_popen.return_value.returncode = 0
        backup_opt = commons.BackupOpt1()
        backup_opt.command = 'echo test'
        jobs.ExecJob(backup_opt, backup_opt.storage).execute()

    def test_execute_raise(self):
        self.popen = patch('freezer.utils.exec_cmd.subprocess.Popen')
        self.mock_popen = self.popen.start()
        self.mock_popen.return_value.returncode = 1
        backup_opt = commons.BackupOpt1()
        backup_opt.command = 'echo test'
        job = jobs.ExecJob(backup_opt, backup_opt.storage)
        self.assertRaises(Exception, job.execute)  # noqa



# (c) Copyright 2014,2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Freezer Backup modes related functions
"""

from freezer.snapshot import lvm
from freezer.snapshot import vss
from freezer.utils import winutils


def snapshot_create(backup_opt_dict):
    """
    Calls the code to take fs snapshots, depending on the platform

    :param backup_opt_dict:
    :return: boolean value, True if snapshot has been taken, false otherwise
    """
    if not backup_opt_dict.snapshot:
        return False

    if winutils.is_windows():
        if backup_opt_dict.snapshot:
            # Create a shadow copy.
            backup_opt_dict.shadow_path, backup_opt_dict.shadow = \
                vss.vss_create_shadow_copy(backup_opt_dict.windows_volume)

            backup_opt_dict.path_to_backup = winutils.use_shadow(
                backup_opt_dict.path_to_backup,
                backup_opt_dict.windows_volume)
            return True
        return False
    else:
        return lvm.lvm_snap(backup_opt_dict)


def snapshot_remove(backup_opt_dict, shadow, windows_volume):
    if winutils.is_windows():
        # Delete the shadow copy after the backup
        vss.vss_delete_shadow_copy(shadow, windows_volume)
    else:
        # Unmount and remove lvm snapshot volume
        lvm.lvm_snap_remove(backup_opt_dict)



# (c) Copyright 2014,2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import errno
import os
import stat

import paramiko

from freezer.storage import fslike
from freezer.utils import utils

CHUNK_SIZE = 32768


class SshStorage(fslike.FsLikeStorage):
    """
    :type ftp: paramiko.SFTPClient
    """
    _type = 'ssh'

    def __init__(self, storage_path, remote_username, remote_ip, port,
                 max_segment_size, ssh_key_path=None, remote_pwd=None):
        """
            :param storage_path: directory of storage
            :type storage_path: str
            :return:
            """
        self.ssh_key_path = ssh_key_path
        self.remote_pwd = remote_pwd
        self.remote_username = remote_username
        self.remote_ip = remote_ip
        self.port = port
        self.ssh = None
        self.ftp = None
        self._validate()
        self.init()
        super(SshStorage, self).__init__(
            storage_path=storage_path,
            max_segment_size=max_segment_size)

    def _validate(self):
        """
        Validates if all parameters required to ssh are available.
        :return: True or raises ValueError
        """
        if not self.remote_ip:
            raise ValueError('Please provide --ssh-host value.')
        elif not self.remote_username:
            raise ValueError('Please provide --ssh-username value.')
        elif self.remote_pwd:
            return True
        elif not self.ssh_key_path:
            raise ValueError('Please provide path to ssh key using '
                             '--ssh-key argument')
        elif not os.path.exists(self.ssh_key_path):
            raise ValueError('1.The {0} is required, please ensure '
                             'it exists and ensure backup node can '
                             'login to {1} as user {2} '
                             'without password.\n'
                             '2. --ssh-password argument is required and '
                             'ensure backup node can login to {1} '
                             'as user {2} with password.'
                             .format(self.ssh_key_path, self.remote_ip,
                                     self.remote_username))
        return True

    def init(self):
        ssh = paramiko.SSHClient()
        ssh.load_system_host_keys()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        if self.remote_pwd:
            ssh.connect(self.remote_ip, username=self.remote_username,
                        password=self.remote_pwd, port=self.port)
        else:
            ssh.connect(self.remote_ip, username=self.remote_username,
                        key_filename=self.ssh_key_path, port=self.port)

        # we should keep link to ssh to prevent garbage collection
        self.ssh = ssh
        self.ftp = self.ssh.open_sftp()

    def _is_dir(self, check_dir):
        return stat.S_IFMT(self.ftp.stat(check_dir).st_mode) == stat.S_IFDIR

    def rmtree(self, path):
        files = self.ftp.listdir(path=path)
        for f in files:
            filepath = utils.path_join(path, f)
            if self._is_dir(filepath):
                self.rmtree(filepath)
            else:
                self.ftp.remove(filepath)
        self.ftp.rmdir(path)

    def create_dirs(self, path):
        """Change to this directory, recursively making new folders if needed.
        Returns True if any folders were created."""
        if path == '/':
            # absolute path so change directory to root
            self.ftp.chdir('/')
            return
        if path == '':
            # top-level relative directory must exist
            return
        try:
            self.ftp.chdir(path)  # sub-directory exists
        except IOError:
            dirname, basename = os.path.split(path.rstrip('/'))
            self.create_dirs(dirname)  # make parent directories
            self.ftp.mkdir(basename)  # sub-directory missing, so created it
            self.ftp.chdir(basename)
            return True

    def get_file(self, from_path, to_path):
        if not self.ssh.get_transport().is_alive():
            self.init()
        self.ftp.get(from_path, to_path)

    def put_file(self, from_path, to_path):
        self.ftp.put(from_path, to_path)

    def listdir(self, directory):
        try:
            # paramiko SFTPClient.listdir_attr returns
            # directories in arbitarary order, so we should
            # sort results of this command
            res = self.ftp.listdir(directory)
            return sorted(res)
        except IOError as e:
            if e.errno == errno.ENOENT:
                return list()
            else:
                raise

    def open(self, filename, mode):
        return self.ftp.open(filename, mode=mode,
                             bufsize=self.max_segment_size)

    def read_metadata_file(self, path):
        file_stats = self.ftp.stat(path)
        file_size = file_stats.st_size
        data = ""
        received_size = 0
        with self.open(path, 'r') as reader:
            reader.prefetch(file_size)
            chunk = reader.read(CHUNK_SIZE)
            while chunk:
                received_size += len(chunk)
                data += chunk
                chunk = reader.read(CHUNK_SIZE)
        if file_size != received_size:
            raise IOError('Size mismatch: expected {} received {}'.format(
                          file_size, received_size))
        return data

    def backup_blocks(self, backup):
        self.init()  # should recreate ssh for new process
        return super(SshStorage, self).backup_blocks(backup)



#!/usr/bin/env python
#
# Copyright 2015 Hewlett-Packard
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from distutils import spawn
import sys
import threading
import time

from apscheduler.schedulers import background
from freezerclient import utils as client_utils
from oslo_config import cfg
from oslo_log import log

from freezer.scheduler import arguments
from freezer.scheduler import scheduler_job
from freezer.scheduler import utils
from freezer.utils import utils as freezer_utils
from freezer.utils import winutils

if winutils.is_windows():
    from freezer.scheduler import win_daemon
else:
    from freezer.scheduler import daemon as linux_daemon

CONF = cfg.CONF
LOG = log.getLogger(__name__)


class FreezerScheduler(object):
    def __init__(self, apiclient, interval, job_path, concurrent_jobs=1):
        # config_manager
        self.client = apiclient
        self.freezerc_executable = spawn.find_executable('freezer-agent')
        if self.freezerc_executable is None:
            # Needed in the case of a non-activated virtualenv
            self.freezerc_executable = spawn.find_executable(
                'freezer-agent', path=':'.join(sys.path))
        LOG.debug('Freezer-agent found at {0}'
                  .format(self.freezerc_executable))
        self.job_path = job_path
        self._client = None
        self.lock = threading.Lock()
        job_defaults = {
            'coalesce': True,
            'max_instances': 1
        }
        executors = {
            'default': {'type': 'threadpool', 'max_workers': 1},
            'threadpool': {'type': 'threadpool',
                           'max_workers': concurrent_jobs}
        }
        self.scheduler = background.BackgroundScheduler(
            job_defaults=job_defaults,
            executors=executors)

        if self.client:
            self.scheduler.add_job(self.poll, 'interval',
                                   seconds=interval, id='api_poll',
                                   executor='default')

        self.add_job = self.scheduler.add_job
        self.remove_job = self.scheduler.remove_job
        self.jobs = {}

    def get_jobs(self):
        if self.client:
            job_doc_list = utils.get_active_jobs_from_api(self.client)
            try:
                utils.save_jobs_to_disk(job_doc_list, self.job_path)
            except Exception as e:
                LOG.error('Unable to save jobs to {0}. '
                          '{1}'.format(self.job_path, e))
            return job_doc_list
        else:
            return utils.get_jobs_from_disk(self.job_path)

    def start_session(self, session_id, job_id, session_tag):
        if self.client:
            return self.client.sessions.start_session(session_id,
                                                      job_id,
                                                      session_tag)
        else:
            raise Exception("Unable to start session: api not in use.")

    def end_session(self, session_id, job_id, session_tag, result):
        if self.client:
            return self.client.sessions.end_session(session_id,
                                                    job_id,
                                                    session_tag,
                                                    result)
        else:
            raise Exception("Unable to end session: api not in use.")

    def upload_metadata(self, metadata_doc):
        if self.client:
            self.client.backups.create(metadata_doc)

    def start(self):
        utils.do_register(self.client)
        self.poll()
        self.scheduler.start()
        try:
            while True:
                # Due to the new Background scheduler nature, we need to keep
                # the main thread alive.
                time.sleep(1)
        except (KeyboardInterrupt, SystemExit):
            # Not strictly necessary if daemonic mode is enabled but
            # should be done if possible
            self.scheduler.shutdown(wait=False)

    def update_job(self, job_id, job_doc):
        if self.client:
            try:
                return self.client.jobs.update(job_id, job_doc)
            except Exception as e:
                LOG.error("Job update error: {0}".format(e))

    def update_job_schedule(self, job_id, job_schedule):
        """
        Pushes to the API the updates the job_schedule information
        of the job_doc

        :param job_id: id of the job to modify
        :param job_schedule: dict containing the job_scheduler information
        :return: None
        """
        doc = {'job_schedule': job_schedule}
        self.update_job(job_id, doc)

    def update_job_status(self, job_id, status):
        doc = {'job_schedule': {'status': status}}
        self.update_job(job_id, doc)

    def is_scheduled(self, job_id):
        return self.scheduler.get_job(job_id) is not None

    def create_job(self, job_doc):
        job = scheduler_job.Job.create(self, self.freezerc_executable, job_doc)
        self.jobs[job.id] = job
        LOG.info("Created job {0}".format(job.id))
        return job

    def poll(self):
        try:
            work_job_doc_list = self.get_jobs()
        except Exception as e:
            LOG.error("Unable to get jobs: {0}".format(e))
            return

        work_job_id_list = []

        # create job if necessary, then let it process its events
        for job_doc in work_job_doc_list:
            job_id = job_doc['job_id']
            work_job_id_list.append(job_id)
            job = self.jobs.get(job_id, None) or self.create_job(job_doc)

            # check for abort status
            if job_doc['job_schedule']['event'] == 'abort':
                pid = int(job_doc['job_schedule']['current_pid'])
                utils.terminate_subprocess(pid, 'freezer-agent')

            job.process_event(job_doc)

        # request removal of any job that has been removed in the api
        for job_id, job in self.jobs.items():
            if job_id not in work_job_id_list:
                job.remove()

        remove_list = [job_id for job_id, job in self.jobs.items()
                       if job.can_be_removed()]

        for k in remove_list:
            self.jobs.pop(k)

    def stop(self):
        sys.exit()

    def reload(self):
        LOG.warning("reload not supported")


def main():
    possible_actions = ['start', 'stop', 'restart', 'status', 'reload']

    arguments.parse_args(possible_actions)
    arguments.setup_logging()

    if CONF.action not in possible_actions:
        CONF.print_help()
        return 65  # os.EX_DATAERR

    apiclient = None
    if CONF.no_api is False:
        try:
            opts = client_utils.Namespace({})
            opts.opts = CONF
            if CONF.enable_v1_api:
                apiclient = client_utils.get_client_instance(opts=opts,
                                                             api_version='1')
            else:
                apiclient = client_utils.get_client_instance(opts=opts,
                                                             api_version='2')

            if CONF.client_id:
                apiclient.client_id = CONF.client_id
        except Exception as e:
            LOG.error(e)
            print(e)
            sys.exit(1)
    else:
        if winutils.is_windows():
            print("--no-api mode is not available on windows")
            return 69  # os.EX_UNAVAILABLE

    freezer_utils.create_dir(CONF.jobs_dir, do_log=False)
    freezer_scheduler = FreezerScheduler(apiclient=apiclient,
                                         interval=int(CONF.interval),
                                         job_path=CONF.jobs_dir,
                                         concurrent_jobs=CONF.concurrent_jobs)

    if CONF.no_daemon:
        print('Freezer Scheduler running in no-daemon mode')
        LOG.debug('Freezer Scheduler running in no-daemon mode')
        if winutils.is_windows():
            daemon = win_daemon.NoDaemon(daemonizable=freezer_scheduler)
        else:
            daemon = linux_daemon.NoDaemon(daemonizable=freezer_scheduler)
    else:
        if winutils.is_windows():
            daemon = win_daemon.Daemon(daemonizable=freezer_scheduler,
                                       interval=int(CONF.interval),
                                       job_path=CONF.jobs_dir,
                                       insecure=CONF.insecure,
                                       concurrent_jobs=CONF.concurrent_jobs)
        else:
            daemon = linux_daemon.Daemon(daemonizable=freezer_scheduler)

    if CONF.action == 'start':
        daemon.start()
    elif CONF.action == 'stop':
        daemon.stop()
    elif CONF.action == 'restart':
        daemon.restart()
    elif CONF.action == 'reload':
        daemon.reload()
    elif CONF.action == 'status':
        daemon.status()

    # os.RETURN_CODES are only available to posix like systems, on windows
    # we need to translate the code to an actual number which is the equivalent
    return 0  # os.EX_OK


if __name__ == '__main__':
    sys.exit(main())



# flake8: noqa
# The MIT License (MIT)
#
# Copyright (c) 2014 Stephan Schultchen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# -*- coding: utf-8 -*-
"""
Simple PidFile Module for a pep3143 daemon implementation.

"""
__author__ = 'schlitzer'


import atexit
import fcntl
import os


class PidFile(object):
    """
    PidFile implementation for PEP 3143 Daemon.

    This Class can also be used with pythons 'with'
    statement.

    :param pidfile:
        filename to be used as pidfile, including path
    :type pidfile: str
    """

    def __init__(self, pidfile):
        """
        Create a new instance
        """
        self._pidfile = pidfile
        self.pidfile = None

    def __enter__(self):
        self.acquire()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        if exc_type is not None:
            self.release()
            return False
        self.release()
        return True

    def acquire(self):
        """Acquire the pidfile.

        Create the pidfile, lock it, write the pid into it
        and register the release with atexit.


        :return: None
        :raise: SystemExit
        """
        try:
            pidfile = open(self._pidfile, "a")
        except IOError as err:
            raise SystemExit(err)
        try:
            fcntl.flock(pidfile.fileno(), fcntl.LOCK_EX | fcntl.LOCK_NB)
        except IOError:
            raise SystemExit('Already running according to ' + self._pidfile)
        pidfile.seek(0)
        pidfile.truncate()
        pidfile.write(str(os.getpid()) + '\n')
        pidfile.flush()
        self.pidfile = pidfile
        atexit.register(self.release)

    def release(self):
        """Release the pidfile.

        Close and delete the Pidfile.


        :return: None
        """
        try:
            self.pidfile.close()
            os.remove(self._pidfile)
        except OSError as err:
            if err.errno != 2:
                raise



# (c) Copyright 2016 Hewlett-Packard Development Enterprise, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
from oslo_config import cfg
from oslo_log import log
from oslo_utils import importutils

from freezer.exceptions import engine as engine_exceptions

CONF = cfg.CONF
LOG = log.getLogger(__name__)


class EngineManager(object):
    """
    EngineManager, this class handles engines.
        Lists all available engines
        Checks if an engine does exists or not
        Load an engine
    """
    def __init__(self):
        """
        This function does some initialization for the engine manager and
        register some variables from the CONF variable into the local class
        """
        self.engine_name = CONF.engine_name
        self.engine_store = os.path.abspath(os.path.dirname(__file__))
        self.engines = self.list_engines()

    def load_engine(self, **kwargs):
        """ Check if the engine exists or not. If not raise EngineNotFound
        Error. If the engine exists then try to get an instance of this engine.
        """
        if not self._check_engine_exists():
            raise engine_exceptions.EngineNotFound(
                "Engine {0} not found".format(self.engine_name)
            )
        return importutils.import_object(
            "freezer.engine.{0}.{0}.{1}Engine".format(
                self.engine_name,
                self.engine_name.capitalize()
            ),
            **kwargs
        )

    def _check_engine_exists(self):
        if self.engine_name not in self.engines:
            return False
        return True

    def list_engines(self):
        """
        Lists all engines in the engine directory
        :return:
        """
        engines = [
            name for name in os.listdir(self.engine_store) if
            os.path.isdir(os.path.join(self.engine_store, name))
        ]
        return engines



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Freezer rsync incremental engine
"""

import fnmatch
import getpass
import grp
import io
import os
import pwd
import queue
import shutil
import stat
import sys
import threading

import msgpack
from oslo_log import log


from freezer.engine import engine
from freezer.engine.rsyncv2 import pyrsync
from freezer.utils import compress
from freezer.utils import crypt
from freezer.utils import winutils

LOG = log.getLogger(__name__)

# Version of the meta data structure format
RSYNC_DATA_STRUCT_VERSION = 2


class Rsyncv2Engine(engine.BackupEngine):
    def __init__(self, **kwargs):
        self.compression_algo = kwargs.get('compression')
        self.encrypt_pass_file = kwargs.get('encrypt_key', None)
        self.dereference_symlink = kwargs.get('symlinks')
        self.exclude = kwargs.get('exclude')
        self.storage = kwargs.get('storage')
        self.is_windows = winutils.is_windows()
        self.dry_run = kwargs.get('dry_run', False)
        self.max_segment_size = kwargs.get('max_segment_size')
        self.rsync_block_size = kwargs.get('rsync_block_size')
        self.fixed_blocks = 0
        self.modified_blocks = 0
        super(Rsyncv2Engine, self).__init__(storage=kwargs.get('storage'))

    @property
    def name(self):
        return "rsync"

    def metadata(self, *args):
        return {
            "engine_name": self.name,
            "compression": self.compression_algo,
            "rsync_block_size": self.rsync_block_size,
            # the encrypt_pass_file might be key content so we need to convert
            # to boolean
            "encryption": bool(self.encrypt_pass_file)
        }

    def backup_data(self, backup_path, manifest_path):
        """Execute backup using rsync algorithm.

        If an existing rsync meta data for file is available - the backup
        will be incremental, otherwise will be executed a level 0 backup.

        :param backup_path: Path to backup
        :param manifest_path: Path to backup metadata
        """

        LOG.info('Starting Rsync engine backup stream')
        LOG.info('Recursively archiving and compressing files '
                 'from {}'.format(os.getcwd()))

        file_read_limit = 0
        data_chunk = b''
        max_seg_size = self.max_segment_size

        # Initialize objects for compressing and encrypting data
        compressor = compress.Compressor(self.compression_algo)
        cipher = None
        if self.encrypt_pass_file:
            cipher = crypt.AESEncrypt(self.encrypt_pass_file)
            yield cipher.generate_header()

        write_queue = queue.Queue(maxsize=2)

        # Create thread for compute file signatures and read data
        t_get_sign_delta = threading.Thread(target=self.get_sign_delta,
                                            args=(
                                                backup_path, manifest_path,
                                                write_queue))
        t_get_sign_delta.daemon = True
        t_get_sign_delta.start()

        # Get backup data from queue
        while True:
            file_block = write_queue.get()

            if file_block is False:
                break

            block_len = len(file_block)
            if block_len == 0:
                continue

            data_chunk += file_block
            file_read_limit += block_len
            if file_read_limit >= max_seg_size:
                yield self._process_backup_data(data_chunk, compressor, cipher)
                data_chunk = b''
                file_read_limit = 0

        flushed_data = self._flush_backup_data(data_chunk, compressor, cipher)

        # Upload segments smaller then max_seg_size
        if len(flushed_data) < max_seg_size:
            yield flushed_data

        # Rejoining thread
        t_get_sign_delta.join()

        LOG.info("Rsync engine backup stream completed")

    @staticmethod
    def _flush_backup_data(data_chunk, compressor, cipher):
        flushed_data = b''
        if data_chunk:
            flushed_data += compressor.compress(data_chunk)

        flushed_data += compressor.flush()
        if flushed_data and cipher:
            flushed_data = cipher.encrypt(flushed_data)

        return flushed_data

    def restore_level(self, restore_path, read_pipe, backup, except_queue):
        """Restore the provided backup into restore_abs_path.

        Decrypt backup content if encrypted.
        Freezer rsync header data structure::

            [ {
                'path': '' (path to file),
                'inode': {
                    'mode': st_mode,
                    'dev': st_dev,
                    'uname': username,
                    'gname': groupname,
                    'atime': st_atime,
                    'mtime': st_mtime,
                    'size': st_size
                } (optional if file removed),
               'lname': 'link_name' (optional if symlink),
               'prev_name': '' (optional if renamed),
               'new_level': True (optional if incremental),
               'deleted': True (optional if removed),
               'deltas': len_of_blocks, [modified blocks] (if patch)
              },
              ...
            ]

        :param restore_path: Path where to restore file(s)
        :param read_pipe: ackup data
        :param backup: Backup info
        :param except_queue: Queue for exceptions
        """

        try:
            metadata = backup.metadata()
            if (not self.encrypt_pass_file and
                    metadata.get("encryption", False)):
                raise Exception("Cannot restore encrypted backup without key")

            self.compression_algo = metadata.get('compression',
                                                 self.compression_algo)

            if not os.path.exists(restore_path):
                raise ValueError(
                    'Provided restore path does not exist: {0}'.format(
                        restore_path))

            if self.dry_run:
                restore_path = '/dev/null'

            data_gen = self._restore_data(read_pipe)

            try:
                data_stream = data_gen.next()
                files_meta, data_stream = self._load_files_meta(data_stream,
                                                                data_gen)

                for fm in files_meta:
                    data_stream = self._restore_file(
                        fm, restore_path, data_stream, data_gen, backup.level)
            except StopIteration:
                LOG.info('Rsync restore process completed')
        except Exception as e:
            LOG.exception(e)
            except_queue.put(e)
            raise

    @staticmethod
    def _load_files_meta(data_stream, data_gen):
        files_meta = []
        while True:
            try:
                files_meta = msgpack.load(data_stream)
            except msgpack.ExtraData as e:
                files_meta = e.unpacked
                data_stream = io.BytesIO(e.extra)
                break
            except msgpack.OutOfData:
                data_stream.write(data_gen.next().read())
                data_stream.seek(0)
            else:
                break
        return files_meta, data_stream

    @staticmethod
    def _remove_file(file_abs_path):
        try:
            if os.path.isdir(file_abs_path):
                shutil.rmtree(file_abs_path)
            else:
                os.unlink(file_abs_path)
        except Exception as e:
            LOG.warning('[*] File or directory unlink error {}'.format(e))

    def _restore_file(self, file_meta, restore_path, data_stream, data_gen,
                      backup_level):
        file_abs_path = os.path.join(restore_path, file_meta['path'])

        inode = file_meta.get('inode', {})
        file_mode = inode.get('mode')

        if os.path.exists(file_abs_path):
            if backup_level == 0:
                self._remove_file(file_abs_path)
            else:
                if file_meta.get('deleted'):
                    self._remove_file(file_abs_path)
                    return data_stream
                elif file_meta.get('new_level') and not stat.S_ISREG(
                        file_mode):
                    self._set_inode(file_abs_path, inode)
                    return data_stream

        if not file_mode:
            return data_stream

        if stat.S_ISREG(file_mode):
            data_stream = self._restore_reg_file(file_abs_path, file_meta,
                                                 data_gen, data_stream)

        elif stat.S_ISDIR(file_mode):
            try:
                os.makedirs(file_abs_path, file_mode)
            except (OSError, IOError) as error:
                LOG.warning(
                    'Directory {0} creation error: {1}'.format(
                        file_abs_path, error))

        elif stat.S_ISBLK(file_mode):
            try:
                self._make_dev_file(file_abs_path, file_meta['dev'], file_mode)
            except (OSError, IOError) as error:
                LOG.warning(
                    'Block file {0} creation error: {1}'.format(
                        file_abs_path, error))

        elif stat.S_ISCHR(file_mode):
            try:
                self._make_dev_file(file_abs_path, file_meta['dev'], file_mode)
            except (OSError, IOError) as error:
                LOG.warning(
                    'Character file {0} creation error: {1}'.format(
                        file_abs_path, error))

        elif stat.S_ISFIFO(file_mode):
            try:
                os.mkfifo(file_abs_path)
            except (OSError, IOError) as error:
                LOG.warning(
                    'FIFO or Pipe file {0} creation error: {1}'.format(
                        file_abs_path, error))

        elif stat.S_ISLNK(file_mode):
            try:
                os.symlink(file_meta.get('lname', ''), file_abs_path)
            except (OSError, IOError) as error:
                LOG.warning('Link file {0} creation error: {1}'.format(
                    file_abs_path, error))

        if not stat.S_ISLNK(file_mode):
            self._set_inode(file_abs_path, inode)

        return data_stream

    @staticmethod
    def _make_dev_file(file_abs_path, dev, mode):
        devmajor = os.major(dev)
        devminor = os.minor(dev)
        new_dev = os.makedev(devmajor, devminor)
        os.mknod(file_abs_path, mode, new_dev)

    @staticmethod
    def _create_reg_file(path, size, data_gen, data_stream):
        with open(path, 'wb') as fd:
            while fd.tell() < size:
                data = data_stream.read(size - fd.tell())
                if not data:
                    data_stream = data_gen.next()
                    continue
                fd.write(data)

        return data_stream

    def _restore_data(self, read_pipe):
        try:
            data_chunk = read_pipe.recv_bytes()
            decompressor = compress.Decompressor(self.compression_algo)
            decryptor = None

            if self.encrypt_pass_file:
                decryptor = crypt.AESDecrypt(self.encrypt_pass_file,
                                             data_chunk[:crypt.BS])
                data_chunk = data_chunk[crypt.BS:]

            while True:
                try:
                    data_chunk = self._process_restore_data(
                        data_chunk, decompressor, decryptor)
                except Exception:
                    data_chunk += read_pipe.recv_bytes()
                    continue
                if data_chunk:
                    yield io.BytesIO(data_chunk)
                data_chunk = read_pipe.recv_bytes()

        except EOFError:
            LOG.info("[*] EOF from pipe. Flushing buffer.")
            data_chunk = decompressor.flush()
            if data_chunk:
                yield io.BytesIO(data_chunk)

    @staticmethod
    def _process_backup_data(data, compressor, encryptor, do_compress=True):
        """Compresses and encrypts provided data according to args"""

        if do_compress:
            data = compressor.compress(data)

        if encryptor:
            data = encryptor.encrypt(data)
        return data

    @staticmethod
    def _process_restore_data(data, decompressor, decryptor):
        """Decrypts and decompresses provided data according to args"""

        if decryptor:
            data = decryptor.decrypt(data)

        data = decompressor.decompress(data)
        return data

    def _get_deltas_info(self, file_name, old_file_meta):
        len_deltas = 0
        old_signature = old_file_meta['signature']
        rsync_bs = self.rsync_block_size

        # Get changed blocks index only
        index = 0
        modified_blocks = []
        with open(file_name, 'rb') as fd:
            for block_index in pyrsync.rsyncdelta_fast(
                    fd,
                    (old_signature[0], old_signature[1]),
                    rsync_bs):

                if not isinstance(block_index, int):
                    block_len = len(block_index)
                    if block_len > 0:
                        len_deltas += block_len
                        modified_blocks.append(index)
                        self.modified_blocks += 1
                else:
                    self.fixed_blocks += 1

                index += 1

        return len_deltas, modified_blocks

    def _backup_deltas(self, file_header, write_queue):
        _, modified_blocks = file_header['deltas']
        rsync_bs = self.rsync_block_size
        with open(file_header['path'], 'rb') as fd:
            for block_index in modified_blocks:
                offset = block_index * rsync_bs
                fd.seek(offset)
                data_block = fd.read(rsync_bs)
                write_queue.put(data_block)

    @staticmethod
    def _is_file_modified(old_inode, inode):
        """Check for changes on inode or file data

        :param old_inode: meta data of the previous backup execution
        :param inode: meta data of the current backup execution
        :return: True if the file changed, False otherwise
        """

        # Check if new ctime/mtime is different from the current one
        file_change_flag = None
        if old_inode['mtime'] != inode['mtime']:
            file_change_flag = True
        elif old_inode['ctime'] != inode['ctime']:
            file_change_flag = True

        return file_change_flag

    def _patch_reg_file(self, file_path, size, data_stream, data_gen,
                        deltas_info):
        len_deltas, modified_blocks = deltas_info
        rsync_bs = self.rsync_block_size
        if len_deltas:
            reminder = len_deltas % rsync_bs
            last_block = modified_blocks.pop()
            # Get all the block index offset from
            with open(file_path, 'rb+') as fd:
                for block_index in modified_blocks:
                    data_stream = self._patch_block(
                        fd, block_index, data_stream, data_gen,
                        rsync_bs, rsync_bs)

                data_stream = self._patch_block(
                    fd, last_block, data_stream, data_gen,
                    reminder if reminder else rsync_bs, rsync_bs)

                fd.truncate(size)

        return data_stream

    @staticmethod
    def _patch_block(fd, block_index, data_stream, data_gen, size, bs):
        offset = block_index * bs
        fd.seek(offset)

        data_stream_pos = data_stream.tell()
        while (data_stream.len - data_stream.tell()) < size:
            data_stream.write(data_gen.next().read())
            data_stream.flush()
            data_stream.seek(data_stream_pos)

        fd.write(data_stream.read(size))
        return data_stream

    def _restore_reg_file(self, file_path, file_meta, data_gen, data_chunk):
        """Create the regular file and write data on it.

        :param size:
        :param file_path:
        :param file_meta:
        :param data_gen:
        :param data_chunk:
        :return:
        """

        new_level = file_meta.get('new_level', False)
        deltas = file_meta.get('deltas')
        size = file_meta['inode']['size']
        if new_level and deltas:
            return self._patch_reg_file(file_path, size, data_chunk,
                                        data_gen, deltas)
        else:
            return self._create_reg_file(file_path, size,
                                         data_gen, data_chunk)

    @staticmethod
    def _set_inode(file_path, inode):
        """Set the file inode fields according to the provided args.

        :param file_path:
        :param inode:
        :return:
        """

        try:
            set_uid = pwd.getpwnam(inode['uname']).pw_uid
            set_gid = grp.getgrnam(inode['gname']).gr_gid
        except (IOError, OSError):
            try:
                set_uid = pwd.getpwnam(getpass.getuser()).pw_uid
                set_gid = grp.getgrnam(getpass.getuser()).gr_gid
            except (OSError, IOError) as err:
                raise Exception(err)
        try:
            os.chown(file_path, set_uid, set_gid)
            os.chmod(file_path, inode['mode'])
            os.utime(file_path, (inode['atime'], inode['mtime']))
        except (OSError, IOError):
            LOG.warning(
                '[*] Unable to set inode info for {}'.format(file_path))

    @staticmethod
    def _parse_file_stat(os_stat):
        header_meta = {
            'mode': os_stat.st_mode,
            'dev': os_stat.st_dev,
            'uname': pwd.getpwuid(os_stat.st_uid).pw_name,
            'gname': grp.getgrgid(os_stat.st_gid).gr_name,
            'atime': os_stat.st_atime,
            'mtime': os_stat.st_mtime,
            'size': os_stat.st_size
        }

        incremental_meta = {
            'mode': os_stat.st_mode,
            'ctime': os_stat.st_ctime,
            'mtime': os_stat.st_mtime
        }

        return header_meta, incremental_meta

    def _get_file_stat(self, rel_path):
        """Generate file meta data from file path.

        Return the meta data as a two dicts: header and incremental

        :param rel_path: related file path
        :return: file meta as a two dicts
        """

        # Get file inode information
        try:
            os_stat = os.lstat(rel_path)
        except (OSError, IOError) as error:
            raise Exception('[*] Error on file stat: {}'.format(error))

        return self._parse_file_stat(os_stat)

    def _backup_file(self, file_path, write_queue):
        max_seg_size = self.max_segment_size
        with open(file_path, 'rb') as file_path_fd:
            data_block = file_path_fd.read(max_seg_size)

            while data_block:
                write_queue.put(data_block)
                data_block = file_path_fd.read(max_seg_size)

    @staticmethod
    def _find_same_inode(file_path, old_files):
        """Find same file meta data for given file name.

        Return the same file name in incremental info if file was removed.

        :param file_path: related file path
        :return: the same file name
        """

        file_name = os.path.basename(file_path)
        for fn in iter(old_files.keys()):
            base_name = os.path.basename(fn)
            if fnmatch.fnmatch(base_name, '*' + file_name + '*'):
                return base_name
        return None

    @staticmethod
    def _get_old_file_meta(file_path, file_stat, old_fs_meta_struct):
        old_file_meta = None
        prev_name = None
        if old_fs_meta_struct:
            try:
                old_file_meta = old_fs_meta_struct[file_path]
                new_mode = file_stat['mode']
                old_mode = old_file_meta['mode']
                if new_mode != old_mode:
                    old_file_meta = None
            except KeyError:
                pass

        return old_file_meta, prev_name

    def _prepare_file_info(self, file_path, old_fs_meta_struct):
        file_stat, file_meta = self._get_file_stat(file_path)
        file_mode = file_stat['mode']

        if stat.S_ISSOCK(file_mode):
            return None, None

        file_header = {'path': file_path, 'inode': file_stat}

        if stat.S_ISLNK(file_mode):
            file_header['lname'] = os.readlink(file_path)

        old_file_meta, old_name = self._get_old_file_meta(
            file_path, file_stat, old_fs_meta_struct)
        if old_name:
            file_header['prev_name'] = old_name

        if old_file_meta:
            if self._is_file_modified(old_file_meta, file_meta):
                file_header['new_level'] = True
            else:
                return old_file_meta, None

        if not stat.S_ISREG(file_mode):
            return file_meta, file_header

        if old_file_meta:
            len_deltas, mod_blocks = self._get_deltas_info(file_path,
                                                           old_file_meta)
            if len_deltas:
                file_header['deltas'] = (len_deltas, mod_blocks)

        return file_meta, file_header

    def _get_file_meta(self, fn, fs_path, old_fs_meta_struct, files_meta,
                       files_header, counts):
        file_path = os.path.relpath(fn, fs_path)
        header_append = files_header.append
        counts['backup_size_on_disk'] += os.path.getsize(file_path)
        meta, header = self._prepare_file_info(file_path, old_fs_meta_struct)
        if meta:
            files_meta['files'][file_path] = meta
        if header:
            header_append(header)

    def _backup_reg_file(self, backup_meta, write_queue):
        if backup_meta.get('deltas'):
            self._backup_deltas(backup_meta, write_queue)
        else:
            self._backup_file(backup_meta['path'], write_queue)

    def get_sign_delta(self, fs_path, manifest_path, write_queue):
        """Compute the file or fs tree path signatures.

        Return blocks of changed data.

        :param fs_path:
        :param manifest_path:
        :param write_queue:
        :return:
        """

        files_meta = {
            'files': {},
            'platform': sys.platform,
            'abs_backup_path': os.getcwd(),
            'rsync_struct_ver': RSYNC_DATA_STRUCT_VERSION,
            'rsync_block_size': self.rsync_block_size}

        counts = {
            'total_files': 0,
            'total_dirs': 0,
            'backup_size_on_disk': 0,
        }

        # Get old file meta structure or an empty dict if not available
        old_fs_meta_struct, rsync_bs = self.get_fs_meta_struct(manifest_path)
        if rsync_bs and rsync_bs != self.rsync_block_size:
            LOG.warning('[*] Incremental backup will be performed '
                        'with rsync_block_size={}'.format(rsync_bs))
            self.rsync_block_size = rsync_bs

        backup_header = []

        # Grab list of all files and directories
        exclude = self.exclude
        if os.path.isdir(fs_path):
            for dn, dl, fl in os.walk(fs_path):
                for dir_name in dl:
                    self._get_file_meta(os.path.join(dn, dir_name),
                                        fs_path, old_fs_meta_struct,
                                        files_meta, backup_header, counts)
                    counts['total_dirs'] += 1

                if exclude:
                    fl = (fn for fn in fl if not fnmatch.fnmatch(fn, exclude))

                for fn in fl:
                    self._get_file_meta(os.path.join(dn, fn), fs_path,
                                        old_fs_meta_struct, files_meta,
                                        backup_header, counts)
                    counts['total_files'] += 1
        else:
            self._get_file_meta(fs_path, os.getcwd(), old_fs_meta_struct,
                                files_meta, backup_header, counts)
            counts['total_files'] += 1

        # Check for deleted files
        for del_file in (f for f in iter(old_fs_meta_struct.keys()) if
                         f not in files_meta['files']):
            backup_header.append({'path': del_file, 'deleted': True})

        # Write backup header
        write_queue.put(msgpack.dumps(backup_header))

        # Backup reg files
        reg_files = (f for f in backup_header if f.get('inode') and
                     stat.S_ISREG(f['inode']['mode']))

        for reg_file in reg_files:
            self._backup_reg_file(reg_file, write_queue)
            self._compute_checksums(reg_file['path'],
                                    files_meta['files'][reg_file['path']])

        LOG.info("Backup session metrics: {0}".format(counts))
        LOG.info("Count of modified blocks %s, count of fixed blocks %s" % (
            self.modified_blocks, self.fixed_blocks))

        self.write_engine_meta(manifest_path, files_meta)

        # Put False on the queue so it will be terminated on the other side:
        write_queue.put(False)

    def write_engine_meta(self, manifest_path, files_meta):
        # Compress meta data file
        # Write meta data to disk as JSON
        with open(manifest_path, 'wb') as manifest_file:
            cmp_meta = compress.one_shot_compress(
                self.compression_algo, msgpack.dumps(files_meta))
            manifest_file.write(cmp_meta)

    def get_fs_meta_struct(self, fs_meta_path):
        old_files_meta = {}

        if os.path.isfile(fs_meta_path):
            with open(fs_meta_path) as meta_file:
                old_files_meta = msgpack.loads(compress.one_shot_decompress(
                    self.compression_algo, meta_file.read()))

        old_fs_meta_struct = old_files_meta.get('files', {})
        rsync_bs = old_files_meta.get('rsync_block_size')

        return old_fs_meta_struct, rsync_bs

    def _compute_checksums(self, rel_path, file_meta):
        # Files type where the file content can be backed up
        args = (rel_path, self.rsync_block_size)
        file_meta['signature'] = pyrsync.blockchecksums(args)



# Copyright 2015 Hewlett-Packard
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import signal
import socket

from freezerclient import exceptions
from oslo_log import log
from oslo_serialization import jsonutils as json
from oslo_utils import uuidutils
import psutil


LOG = log.getLogger(__name__)

CONFIG_FILE_EXT = '.conf'


def do_register(client, args=None):
    if client:
        client_info = {
            "client_id": client.client_id,
            "hostname": socket.gethostname()
        }
        try:
            client.clients.create(client_info)
        except exceptions.ApiClientException as e:
            if e.status_code == 409:
                print("Client already registered")
            return 73  # os.EX_CANTCREAT
        return 0  # os.EX_OK


def find_config_files(path):
    expanded_path = os.path.expanduser(path)
    if os.path.isfile(expanded_path):
        return [expanded_path]
    file_list = []
    if os.path.isdir(expanded_path):
        for fname in next(os.walk(expanded_path))[2]:
            if CONFIG_FILE_EXT.upper() == os.path.splitext(fname)[1].upper():
                file_list.append('{0}/{1}'.format(expanded_path, fname))
        return file_list
    raise Exception("unable to find job files at the provided path "
                    "{0}".format(path))


def load_doc_from_json_file(fname, debug=False):
    with open(fname, 'rb') as fd:
        try:
            doc = json.load(fd)
        except Exception as e:
            raise Exception("Unable to load conf file. {0}".format(e))
        if debug:
            print("File {0} loaded: ".format(fname))
        return doc


def save_doc_to_json_file(doc, fname, debug=False):
    with open(fname, 'w') as fd:
        json.dump(doc, fd, indent=4)
    if debug:
        print('Saved doc to file: {0}'.format(fname))


def get_jobs_from_disk(path):
    job_doc_list = [
        load_doc_from_json_file(f) for f in find_config_files(path)]
    for job_doc in job_doc_list:
        if job_doc:
            job_doc['job_id'] = job_doc.get('job_id', uuidutils.generate_uuid(
                dashed=False))
    return [x for x in job_doc_list if x]


def save_jobs_to_disk(job_doc_list, path):
    for doc in job_doc_list:
        fname = os.path.normpath('{0}/job_{1}.conf'.
                                 format(path, doc['job_id']))
        save_doc_to_json_file(doc, fname)


def get_active_jobs_from_api(client):
    # might raise
    search = {"match_not": [{"status": "completed"}]}
    job_list, offset = [], 0
    while True:
        jobs = client.jobs.list(limit=10, offset=offset, search=search)
        job_list.extend(jobs)
        if len(jobs) < 10:
            break
        offset += len(jobs)
    return job_list


def terminate_subprocess(pid, name):
    try:
        process = psutil.Process(pid)
        if process.name.startswith(name):
            os.kill(pid, signal.SIGTERM)
        else:
            LOG.warning('The name {} does not match the pid {}'.format(
                name, pid))
    except Exception:
        LOG.debug('Process {} does not exists anymore'.format(pid))



# (c) Copyright 2014,2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_log import log
# PyCharm will not recognize queue. Puts red squiggle line under it. That's OK.
import queue

from freezer.storage import base
from freezer.storage import exceptions
from freezer.utils import streaming

LOG = log.getLogger(__name__)


class MultipleStorage(base.Storage):
    _type = 'multiple'

    def __init__(self, storages):
        """
        :param storages:
        :type storages: list[freezer.storage.base.Storage]
        :return:
        """
        super(MultipleStorage, self).__init__()
        self.storages = storages

    def info(self):
        for s in self.storages:
            s.info()

    def write_backup(self, rich_queue, backup):
        output_queues = [streaming.RichQueue() for x in self.storages]
        except_queues = [queue.Queue() for x in self.storages]
        threads = ([streaming.QueuedThread(storage.write_backup, output_queue,
                    except_queue, kwargs={"backup": backup}) for
                    storage, output_queue, except_queue in
                    zip(self.storages, output_queues, except_queues)])

        for thread in threads:
            thread.daemon = True
            thread.start()

        StorageManager(rich_queue, output_queues).transmit()

        for thread in threads:
            thread.join()

        def handle_exception_queue(except_queue):
            if not except_queue.empty:
                while not except_queue.empty():
                    e = except_queue.get_nowait()
                    LOG.critical('Storage error: {0}'.format(e))
                return True
            else:
                return False

        got_exception = None
        for except_queue in except_queues:
            got_exception = (handle_exception_queue(except_queue) or
                             got_exception)

        if (got_exception):
            raise exceptions.StorageException(
                "Storage error. Failed to backup.")

    def get_level_zero(self,
                       engine,
                       hostname_backup_name,
                       recent_to_date=None):
        backups = (
            [s.get_level_zero(engine, hostname_backup_name, recent_to_date)
             for s in self.storages]
        )
        # flat the list
        return [item for sublist in backups for item in sublist]

    def prepare(self):
        pass

    def put_file(self, from_path, to_path):
        for storage in self.storages:
            storage.put_file(from_path, to_path)

    def put_engine_metadata(self, from_path, backup):
        """

        :param from_path:
        :type backup: freezer.storage.base.Backup
        :param backup:
        :return:
        """
        for storage in self.storages:
            storage.put_engine_metadata(from_path, backup)

    def download_freezer_meta_data(self, backup):
        # TODO(DEKLAN): Need to implement.
        pass

    def get_file(self, from_path, to_path):
        # TODO(DEKLAN): Need to implement.
        pass

    def meta_file_abs_path(self, backup):
        # TODO(DEKLAN): Need to implement.
        pass

    def upload_freezer_meta_data(self, backup, meta_dict):
        # TODO(DEKLAN): Need to implement.
        pass

    def create_dirs(self, path):
        for storage in self.storages:
            storage.create_dirs(path)

    def put_metadata(self,
                     engine_metadata_path,
                     freezer_metadata_path,
                     backup):
        for storage in self.storages:
            storage.put_metadata(engine_metadata_path,
                                 freezer_metadata_path,
                                 backup)


class StorageManager(object):

    def __init__(self, input_queue, output_queues):
        """
        :type input_queue: streaming.RichQueue
        :param input_queue:
        :type output_queues: collections.Iterable[streaming.RichQueue]
        :param output_queues:
        :return:
        """
        self.input_queue = input_queue
        self.output_queues = output_queues
        self.broken_output_queues = set()

    def send_message(self, message, finish=False):
        for output_queue in self.output_queues:
            if output_queue not in self.broken_output_queues:
                try:
                    if finish:
                        output_queue.finish()
                    else:
                        output_queue.put(message)
                except Exception as e:
                    LOG.exception(e)
                    StorageManager.one_fails_all_fail(
                        self.input_queue, self.output_queues)
                    self.broken_output_queues.add(output_queue)

    def transmit(self):
        for message in self.input_queue.get_messages():
            self.send_message(message)
        self.send_message("", True)

    @staticmethod
    def one_fails_all_fail(input_queue, output_queues):
        input_queue.force_stop()
        for output_queue in output_queues:
            output_queue.force_stop()
        raise Exception("All fail")



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import hashlib
import zlib


_BASE = 65521  # largest prime smaller than 65536


def adler32fast(data):
    return zlib.adler32(data) & 0xffffffff


def adler32(data):
    checksum = zlib.adler32(data)
    s2, s1 = (checksum >> 16) & 0xffff, checksum & 0xffff
    return checksum & 0xffffffff, s1, s2


def adler32rolling(removed, new, s1, s2, blocksize=4096):
    r = ord(removed)
    n = ord(new)
    s1 = (s1 + n - r) % _BASE
    s2 = (s2 + s1 - blocksize * r - 1) % _BASE
    return ((s2 << 16) | s1) & 0xffffffff, s1, s2


def blockchecksums(args):
    """
    Returns a list of weak and strong hashes for each block of the
    defined size for the given data stream.
    """
    path, blocksize = args
    weakhashes = []
    stronghashes = []
    weak_append = weakhashes.append
    strong_append = stronghashes.append

    with open(path, 'rb') as instream:
        instream_read = instream.read
        read = instream_read(blocksize)

        while read:
            weak_append(adler32fast(read))
            strong_append(hashlib.sha1(read).hexdigest())
            read = instream_read(blocksize)

    return weakhashes, stronghashes


def rsyncdelta_fast(datastream, remotesignatures, blocksize=4096):
    rem_weak, rem_strong = remotesignatures
    data_block = datastream.read(blocksize)
    index = 0
    while data_block:
        try:
            if adler32fast(data_block) == rem_weak[index] and hashlib.sha1(
                    data_block).hexdigest() == rem_strong[index]:
                yield index
            else:
                yield data_block
        except IndexError:
            yield data_block

        index += 1
        data_block = datastream.read(blocksize)



# (c) Copyright 2014,2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import botocore
import botocore.session
import logging
import requests

from oslo_log import log
from oslo_serialization import jsonutils as json
from requests.packages.urllib3.exceptions import InsecureRequestWarning

from freezer.storage import physical
from freezer.utils import utils

LOG = log.getLogger(__name__)
logging.getLogger('botocore').setLevel(logging.WARNING)
requests.packages.urllib3.disable_warnings(InsecureRequestWarning)


class S3Storage(physical.PhysicalStorage):

    _type = 's3'

    def __init__(self, access_key, secret_key, endpoint, container,
                 max_segment_size, skip_prepare=False):
        """
        :type container: str
        """
        self.access_key = access_key
        self.secret_key = secret_key
        self.endpoint = endpoint
        super(S3Storage, self).__init__(
            storage_path=container,
            max_segment_size=max_segment_size,
            skip_prepare=skip_prepare)
        self.container = container
        storage_info = self.get_storage_info()
        self._bucket_name = storage_info[0]
        self._object_prefix = storage_info[1]

    def get_storage_info(self):
        storage_info = self.storage_path.split('/', 1)
        bucket_name = storage_info[0]
        object_prefix = storage_info[1] if '/' in self.storage_path else ''
        return bucket_name, object_prefix

    def rmtree(self, path):
        split = path.split('/', 1)
        all_s3_objects = self.list_all_objects(
            bucket_name=split[0],
            prefix=split[1]
        )
        for s3_object in all_s3_objects:
            self.get_s3_connection().delete_object(
                Bucket=split[0],
                Key=s3_object['Key']
            )

    def put_file(self, from_path, to_path):
        split = to_path.split('/', 1)
        self.get_s3_connection().put_object(
            Bucket=split[0],
            Key=split[1],
            Body=open(from_path, 'r')
        )

    def get_s3_connection(self):
        """
        :rtype: s3client.Connection
        :return:
        """
        return botocore.session.get_session().create_client(
            's3',
            aws_access_key_id=self.access_key,
            aws_secret_access_key=self.secret_key,
            endpoint_url=self.endpoint
        )

    def prepare(self):
        """
        Check if the provided bucket is already available on S3
        compatible storage. The verification is done by exact matching
        between the provided bucket name and the whole list of bucket
        available for the S3 account.
        """
        bucket_list = \
            [c['Name'] for c in self.get_s3_connection()
                .list_buckets()['Buckets']]
        split = self.storage_path.split('/', 1)
        if split[0] not in bucket_list:
            self.get_s3_connection().create_bucket(Bucket=split[0])

    def info(self):
        # S3 standard interface do not have stats query function
        buckets = self.get_s3_connection().list_buckets()['Buckets']
        ordered_buckets = []
        for bucket in buckets:
            ordered_bucket = {
                'container_name': bucket['Name'],
                'size': '{0} MB'.format(0),
                'objects_count': 0
            }
            ordered_buckets.append(ordered_bucket)
        return ordered_buckets

    def get_file(self, from_path, to_path):
        split = from_path.split('/', 1)
        with open(to_path, 'ab') as obj_fd:
            s3_object = self.get_s3_connection().get_object(
                Bucket=split[0],
                Key=split[1])['Body']
            while True:
                object_trunk = s3_object.read(self.max_segment_size)
                if len(object_trunk) == 0:
                    break
                obj_fd.write(object_trunk)

    def add_stream(self, stream, package_name, headers=None):

        split = package_name.rsplit('/', 1)
        backup_basedir = package_name \
            if self.get_object_prefix() == '' \
            else "{0}/{1}".format(self.get_object_prefix(), package_name)

        backup_basepath = "{0}/{1}".format(backup_basedir, split[0])
        backup_meta_data = "{0}/metadata".format(backup_basedir)

        self.upload_stream(backup_basepath, stream)

        self.get_s3_connection().put_object(
            Bucket=self.get_bucket_name(),
            Body=json.dumps(headers),
            Key=backup_meta_data
        )

    def upload_stream(self, backup_basepath, stream):
        upload_id = self.get_s3_connection().create_multipart_upload(
            Bucket=self.get_bucket_name(),
            Key=backup_basepath
        )['UploadId']
        upload_part_index = 1
        uploaded_parts = []
        try:
            for el in stream:
                response = self.get_s3_connection().upload_part(
                    Body=el,
                    Bucket=self.get_bucket_name(),
                    Key=backup_basepath,
                    PartNumber=upload_part_index,
                    UploadId=upload_id
                )
                uploaded_parts.append({
                    'PartNumber': upload_part_index,
                    'ETag': response['ETag']
                })
                upload_part_index += 1
            # Complete the upload, which requires info on all of the parts
            part_info = {
                'Parts': uploaded_parts
            }

            if not uploaded_parts:
                # currently, not support volume boot instance
                LOG.error(
                    "No part uploaded(not support volume boot instance)"
                )
                raise RuntimeError(
                    'No part uploaded(not support volume boot instance)'
                )

            self.get_s3_connection().complete_multipart_upload(
                Bucket=self.get_bucket_name(),
                Key=backup_basepath,
                MultipartUpload=part_info,
                UploadId=upload_id
            )
        except Exception as e:
            LOG.error("Upload stream to S3 error, so abort it. "
                      "Exception: {0}".format(e))
            self.get_s3_connection().abort_multipart_upload(
                Bucket=self.get_bucket_name(),
                Key=backup_basepath,
                UploadId=upload_id
            )

    def backup_blocks(self, backup):
        """
        :param backup:
        :type backup: freezer.storage.base.Backup
        :return:
        """
        split = backup.data_path.split('/', 1)
        s3_object = self.get_s3_connection().get_object(
            Bucket=split[0],
            Key=split[1]
        )

        return utils.S3ResponseStream(
            data=s3_object['Body'],
            chunk_size=self.max_segment_size
        )

    def write_backup(self, rich_queue, backup):
        """
        Upload object to the remote S3 compatible storage server
        :type rich_queue: freezer.streaming.RichQueue
        :type backup: freezer.storage.base.Backup
        """
        backup = backup.copy(storage=self)
        path = backup.data_path.split('/', 1)[1]
        self.upload_stream(path, rich_queue.get_messages())

    def listdir(self, path):
        """
        :type path: str
        :param path:
        :rtype: collections.Iterable[str]
        """
        try:
            files = set()
            split = path.split('/', 1)
            all_s3_objects = self.list_all_objects(
                bucket_name=split[0],
                prefix=split[1]
            )
            for s3_object in all_s3_objects:
                files.add(s3_object['Key'][len(split[1]):].split('/', 2)[1])
            return files
        except Exception as e:
            LOG.error(e)
            return []

    def create_dirs(self, folder_list):
        pass

    def get_bucket_name(self):
        return self._bucket_name

    def get_object_prefix(self):
        return self._object_prefix

    def get_object(self, bucket_name, key):
        return self.get_s3_connection().get_object(
            Bucket=bucket_name,
            Key=key
        )

    def list_all_objects(self, bucket_name, prefix):
        all_objects = []
        is_truncated = True
        marker = ''
        while is_truncated:
            response = self.get_s3_connection().list_objects(
                Bucket=bucket_name,
                Marker=marker,
                Prefix=prefix
            )

            if 'Contents' not in response:
                break

            all_objects.extend(response['Contents'])
            is_truncated = response['IsTruncated']
            marker = response['Contents'][-1]['Key']
        return all_objects

    def put_object(self, bucket_name, key, data):
        self.get_s3_connection().put_object(
            Bucket=bucket_name,
            Key=key,
            Body=data
        )



# (c) Copyright 2014,2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
from unittest import mock
from unittest.mock import patch

from freezer.snapshot import lvm


class Test_lvm_snap_remove(unittest.TestCase):
    @patch('freezer.snapshot.lvm.os')
    @patch('freezer.snapshot.lvm._umount')
    @patch('freezer.snapshot.lvm._lvremove')
    def test_return_none_on_success(self, mock_lvremove, mock_umount, mock_os):
        backup_opt = mock.Mock()
        backup_opt.lvm_volgroup = 'one'
        backup_opt.lvm_snapname = 'two'
        self.assertIsNone(lvm.lvm_snap_remove(backup_opt))


class Test_lvm_snap(unittest.TestCase):
    @patch('freezer.snapshot.lvm.get_lvm_info')
    @patch('freezer.snapshot.lvm.utils.create_dir')
    def test_with_snapshot_opt_simple_sets_correct_path_and_raises_on_perm(
            self, mock_create_dir, mock_get_lvm_info):
        mock_get_lvm_info.return_value = {
            'volgroup': 'lvm_volgroup',
            'srcvol': 'lvm_device',
            'snap_path': 'snap_path'}

        backup_opt = mock.Mock()
        backup_opt.snapshot = True
        backup_opt.path_to_backup = '/just/a/path'
        backup_opt.lvm_dirmount = '/var/mountpoint'
        backup_opt.lvm_snapperm = 'invalid_value'

        with self.assertRaises(Exception) as cm:  # noqa
            lvm.lvm_snap(backup_opt)
        the_exception = cm.exception
        self.assertIn('Invalid value for option lvm-snap-perm',
                      str(the_exception))

    @patch('freezer.snapshot.lvm.validate_lvm_params')
    @patch('freezer.snapshot.lvm.subprocess.Popen')
    @patch('freezer.snapshot.lvm.get_vol_fs_type')
    @patch('freezer.snapshot.lvm.get_lvm_info')
    @patch('freezer.snapshot.lvm.utils.create_dir')
    def test_ok(self, mock_create_dir, mock_get_lvm_info, mock_get_vol_fs_type,
                mock_popen, mock_validate_lvm_params):
        mock_get_lvm_info.return_value = {
            'volgroup': 'lvm_volgroup',
            'srcvol': 'lvm_device',
            'snap_path': 'snap_path'}
        mock_process = mock.Mock()
        mock_process.communicate.return_value = '', ''
        mock_process.returncode = 0
        mock_popen.return_value = mock_process

        backup_opt = mock.Mock()
        backup_opt.snapshot = True
        backup_opt.path_to_backup = '/just/a/path'
        backup_opt.lvm_dirmount = '/var/mountpoint'
        backup_opt.lvm_snapperm = 'ro'
        backup_opt.lvm_volgroup = ''
        backup_opt.lvm_srcvol = ''
        backup_opt.lvm_snapsize = '1G'

        self.assertTrue(lvm.lvm_snap(backup_opt))

    @patch('freezer.snapshot.lvm.validate_lvm_params')
    @patch('freezer.snapshot.lvm.subprocess.Popen')
    @patch('freezer.snapshot.lvm.get_vol_fs_type')
    @patch('freezer.snapshot.lvm.get_lvm_info')
    @patch('freezer.utils.utils.create_dir')
    def test_snapshot_fails(self, mock_create_dir, mock_get_lvm_info,
                            mock_get_vol_fs_type, mock_popen,
                            mock_validate_lvm_params):
        mock_get_lvm_info.return_value = {
            'volgroup': 'lvm_volgroup',
            'srcvol': 'lvm_device',
            'snap_path': 'snap_path'}
        mock_process = mock.Mock()
        mock_process.communicate.return_value = '', ''
        mock_process.returncode = 1
        mock_popen.return_value = mock_process

        backup_opt = mock.Mock()
        backup_opt.snapshot = True
        backup_opt.path_to_backup = '/just/a/path'
        backup_opt.lvm_dirmount = '/var/mountpoint'
        backup_opt.lvm_snapperm = 'ro'
        backup_opt.lvm_snapsize = '1G'

        with self.assertRaises(Exception) as cm:  # noqa
            lvm.lvm_snap(backup_opt)
        the_exception = cm.exception
        self.assertIn('lvm snapshot creation error', str(the_exception))

    @patch('freezer.snapshot.lvm.lvm_snap_remove')
    @patch('freezer.snapshot.lvm.validate_lvm_params')
    @patch('freezer.snapshot.lvm.subprocess.Popen')
    @patch('freezer.snapshot.lvm.get_vol_fs_type')
    @patch('freezer.snapshot.lvm.get_lvm_info')
    @patch('freezer.utils.utils.create_dir')
    def test_already_mounted(self, mock_create_dir, mock_get_lvm_info,
                             mock_get_vol_fs_type,
                             mock_popen, mock_validate_lvm_params,
                             lvm_snap_remove):
        mock_get_vol_fs_type.return_value = 'xfs'
        mock_get_lvm_info.return_value = {
            'volgroup': 'lvm_volgroup',
            'srcvol': 'lvm_device',
            'snap_path': 'snap_path'}
        mock_process = mock.Mock()
        mock_process.communicate.return_value = '', 'already mounted'
        mock_process.returncode = 0
        mock_popen.return_value = mock_process

        backup_opt = mock.Mock()
        backup_opt.snapshot = True
        backup_opt.path_to_backup = '/just/a/path'
        backup_opt.lvm_dirmount = '/var/mountpoint'
        backup_opt.lvm_snapperm = 'ro'
        backup_opt.lvm_snapsize = '1G'

        self.assertTrue(lvm.lvm_snap(backup_opt))

    @patch('freezer.snapshot.lvm.lvm_snap_remove')
    @patch('freezer.snapshot.lvm.subprocess.Popen')
    @patch('freezer.snapshot.lvm.get_vol_fs_type')
    @patch('freezer.snapshot.lvm.get_lvm_info')
    @patch('freezer.snapshot.lvm.utils.create_dir')
    def test_snapshot_mount_error_raises_Exception(self,
                                                   mock_create_dir,
                                                   mock_get_lvm_info,
                                                   mock_get_vol_fs_type,
                                                   mock_popen,
                                                   mock_lvm_snap_remove):
        mock_get_vol_fs_type.return_value = 'xfs'
        mock_get_lvm_info.return_value = {
            'volgroup': 'lvm_volgroup',
            'srcvol': 'lvm_device',
            'snap_path': 'snap_path'}
        mock_lvcreate_process, mock_mount_process = mock.Mock(), mock.Mock()

        mock_lvcreate_process.communicate.return_value = '', ''
        mock_lvcreate_process.returncode = 0

        mock_mount_process.communicate.return_value = '', 'mount error'
        mock_mount_process.returncode = 1

        mock_popen.side_effect = [mock_lvcreate_process, mock_mount_process]

        backup_opt = mock.Mock()
        backup_opt.snapshot = True
        backup_opt.path_to_backup = '/just/a/path'
        backup_opt.lvm_dirmount = '/var/mountpoint'
        backup_opt.lvm_snapperm = 'ro'
        backup_opt.lvm_snapsize = '1G'

        with self.assertRaises(Exception) as cm:  # noqa
            lvm.lvm_snap(backup_opt)
        the_exception = cm.exception
        self.assertIn('lvm snapshot mounting error', str(the_exception))

        mock_lvm_snap_remove.assert_called_once_with(backup_opt)


# class Test_get_lvm_info(unittest.TestCase):
#     @patch('freezer.snapshot.lvm.lvm_guess')
#     @patch('freezer.snapshot.lvm.utils.get_mount_from_path')
#     def test_using_guess(self, mock_get_mount_from_path, mock_lvm_guess):
#         mock_get_mount_from_path.return_value = ('/home/somedir',
#                                                  'some-snap-path')
#         mock_lvm_guess.return_value = 'vg_test', 'lv_test', 'lvm_device'
#         mounts = (
#             '/dev/mapper/vg_prova-lv_prova_vol1 /home/pippo ext4 rw,'
#             'relatime,data=ordered 0 0'
#         )
#         mocked_open_function = mock.mock_open(read_data=mounts)
#
#         with patch("__builtin__.open", mocked_open_function):
#             res = lvm.get_lvm_info('lvm_auto_snap_value')
#
#         expected_result = {'volgroup': 'vg_test',
#                            'snap_path': 'some-snap-path',
#                            'srcvol': 'lvm_device'}
#         self.assertEqual(res, expected_result)
#
#     @patch('freezer.snapshot.lvm.subprocess.Popen')
#     @patch('freezer.snapshot.lvm.lvm_guess')
#     @patch('freezer.snapshot.lvm.utils.get_mount_from_path')
#     def test_using_mount(self, mock_get_mount_from_path, mock_lvm_guess,
#                          mock_popen):
#         mock_get_mount_from_path.return_value = ('/home/somedir',
#                                                  'some-snap-path')
#         mock_lvm_guess.side_effect = [(None, None, None),
#                                       ('vg_test', 'lv_test', 'lvm_device')]
#         mounts = (
#             '/dev/mapper/vg_prova-lv_prova_vol1 /home/pippo ext4 rw,'
#             'relatime,data=ordered 0 0'
#         )
#         mocked_open_function = mock.mock_open(read_data=mounts)
#         mock_process = mock.Mock()
#         mock_process.returncode = 0
#         mock_popen.return_value = mock_process
#         mock_process.communicate.return_value = '', ''
#
#         with patch("__builtin__.open", mocked_open_function):
#             res = lvm.get_lvm_info('lvm_auto_snap_value')
#
#         expected_result = {'volgroup': 'vg_test',
#                            'snap_path': 'some-snap-path',
#                            'srcvol': 'lvm_device'}
#         self.assertEqual(res, expected_result)
#
#     @patch('freezer.snapshot.lvm.subprocess.Popen')
#     @patch('freezer.snapshot.lvm.lvm_guess')
#     @patch('freezer.snapshot.lvm.utils.get_mount_from_path')
#     def test_raises_Exception_when_info_not_found(
#             self, mock_get_mount_from_path, mock_lvm_guess, mock_popen):
#         mock_get_mount_from_path.return_value = ('/home/somedir',
#                                                  'some-snap-path')
#         mock_lvm_guess.return_value = None, None, None
#         mounts = (
#             '/dev/mapper/vg_prova-lv_prova_vol1 /home/pippo ext4 rw,'
#             'relatime,data=ordered 0 0'
#         )
#         mocked_open_function = mock.mock_open(read_data=mounts)
#         mock_process = mock.Mock()
#         mock_lvm_guess.return_value = None, None, None
#         mock_process.communicate.return_value = '', ''
#         mock_popen.return_value = mock_process
#
#         with patch("__builtin__.open", mocked_open_function):
#             self.assertRaises(Exception, lvm.get_lvm_info,
#                               'lvm_auto_snap_value')


class Test_lvm_guess(unittest.TestCase):
    def test_no_match(self):
        mount_points = []
        mount_point_path = '/home/pippo'
        source = '/proc/mounts'

        res = lvm.lvm_guess(mount_point_path, mount_points, source)

        expected_result = (None, None, None)
        self.assertEqual(res, expected_result)

    def test_unsing_proc_mounts(self):
        mount_points = [
            'rootfs / rootfs rw 0 0\n',
            'sysfs /sys sysfs rw,nosuid,nodev,noexec,relatime 0 0\n',
            'proc /proc proc rw,nosuid,nodev,noexec,relatime 0 0\n',
            'udev /dev devtmpfs rw,relatime,size=2010616k,nr_inodes=502654,'
            'mode=755 0 0\n',
            'devpts /dev/pts devpts rw,nosuid,noexec,relatime,gid=5,mode=620,'
            'ptmxmode=000 0 0\n',
            'tmpfs /run tmpfs rw,nosuid,noexec,relatime,size=404836k,mode=755 '
            '0 0\n',
            '/dev/mapper/fabuntu--vg-root / ext4 rw,relatime,'
            'errors=remount-ro,data=ordered 0 0\n',
            'none /sys/fs/cgroup tmpfs rw,relatime,size=4k,mode=755 0 0\n',
            'none /sys/fs/fuse/connections fusectl rw,relatime 0 0\n',
            'none /sys/kernel/debug debugfs rw,relatime 0 0\n',
            'none /sys/kernel/security securityfs rw,relatime 0 0\n',
            'cgroup /sys/fs/cgroup/cpuset cgroup rw,relatime,cpuset 0 0\n',
            'cgroup /sys/fs/cgroup/cpu cgroup rw,relatime,cpu 0 0\n',
            'cgroup /sys/fs/cgroup/cpuacct cgroup rw,relatime,cpuacct 0 0\n',
            'cgroup /sys/fs/cgroup/memory cgroup rw,relatime,memory 0 0\n',
            'none /run/lock tmpfs rw,nosuid,nodev,noexec,relatime,size=5120k '
            '0 0\n',
            'none /run/shm tmpfs rw,nosuid,nodev,relatime 0 0\n',
            'none /run/user tmpfs rw,nosuid,nodev,noexec,relatime,'
            'size=102400k,mode=755 0 0\n',
            'cgroup /sys/fs/cgroup/devices cgroup rw,relatime,devices 0 0\n',
            'none /sys/fs/pstore pstore rw,relatime 0 0\n',
            'cgroup /sys/fs/cgroup/freezer cgroup rw,relatime,freezer 0 0\n',
            'cgroup /sys/fs/cgroup/blkio cgroup rw,relatime,blkio 0 0\n',
            'cgroup /sys/fs/cgroup/perf_event cgroup rw,relatime,perf_event '
            '0 0\n',
            'cgroup /sys/fs/cgroup/hugetlb cgroup rw,relatime,hugetlb 0 0\n',
            '/dev/sda1 /boot ext2 rw,relatime 0 0\n',
            'systemd /sys/fs/cgroup/systemd cgroup rw,nosuid,nodev,noexec,'
            'relatime,name=systemd 0 0\n',
            '/dev/mapper/vg_prova-lv_prova_vol1 /home/pippo ext4 rw,relatime,'
            'data=ordered 0 0\n'
        ]
        mount_point_path = '/home/pippo'
        source = '/proc/mounts'

        res = lvm.lvm_guess(mount_point_path, mount_points, source)

        expected_result = (
            'vg_prova', 'lv_prova_vol1', '/dev/vg_prova/lv_prova_vol1')
        self.assertEqual(res, expected_result)

    def test_unsing_mount(self):
        mount_points = [
            '/dev/mapper/fabuntu--vg-root on / type ext4 (rw,errors='
            'remount-ro)',
            'proc on /proc type proc (rw,noexec,nosuid,nodev)',
            'sysfs on /sys type sysfs (rw,noexec,nosuid,nodev)',
            'none on /sys/fs/cgroup type tmpfs (rw)',
            'none on /sys/fs/fuse/connections type fusectl (rw)',
            'none on /sys/kernel/debug type debugfs (rw)',
            'none on /sys/kernel/security type securityfs (rw)',
            'udev on /dev type devtmpfs (rw,mode=0755)',
            'devpts on /dev/pts type devpts (rw,noexec,nosuid,gid=5,'
            'mode=0620)',
            'tmpfs on /run type tmpfs (rw,noexec,nosuid,size=10%,mode=0755)',
            'none on /run/lock type tmpfs (rw,noexec,nosuid,nodev,'
            'size=5242880)',
            'none on /run/shm type tmpfs (rw,nosuid,nodev)',
            'none on /run/user type tmpfs (rw,noexec,nosuid,nodev,'
            'size=104857600,mode=0755)',
            'none on /sys/fs/pstore type pstore (rw)',
            'cgroup on /sys/fs/cgroup/cpuset type cgroup (rw,relatime,cpuset)',
            'cgroup on /sys/fs/cgroup/cpu type cgroup (rw,relatime,cpu)',
            'cgroup on /sys/fs/cgroup/cpuacct type cgroup (rw,relatime,'
            'cpuacct)',
            'cgroup on /sys/fs/cgroup/memory type cgroup (rw,relatime,memory)',
            'cgroup on /sys/fs/cgroup/devices type cgroup (rw,relatime,'
            'devices)',
            'cgroup on /sys/fs/cgroup/freezer type cgroup (rw,relatime,'
            'freezer)',
            'cgroup on /sys/fs/cgroup/blkio type cgroup (rw,relatime,blkio)',
            'cgroup on /sys/fs/cgroup/perf_event type cgroup (rw,relatime,'
            'perf_event)',
            'cgroup on /sys/fs/cgroup/hugetlb type cgroup (rw,relatime,'
            'hugetlb)',
            '/dev/sda1 on /boot type ext2 (rw)',
            'systemd on /sys/fs/cgroup/systemd type cgroup (rw,noexec,nosuid,'
            'nodev,none,name=systemd)',
            '/dev/mapper/vg_prova-lv_prova_vol1 on /home/pippo type ext4 (rw)',
            '']
        mount_point_path = '/home/pippo'
        source = 'mount'

        res = lvm.lvm_guess(mount_point_path, mount_points, source)

        expected_result = ('vg_prova', 'lv_prova_vol1',
                           '/dev/vg_prova/lv_prova_vol1')
        self.assertEqual(res, expected_result)


class Test_validate_lvm_params(unittest.TestCase):
    def setUp(self):
        self.backup_opt = mock.Mock()
        self.backup_opt.lvm_snapperm = 'ro'
        self.backup_opt.path_to_backup = '/path/to/backup'
        self.backup_opt.lvm_srcvol = '/lvm/srcvol'
        self.backup_opt.lvm_volgroup = 'vg_burger'

    def test_return_true_on_lvm_configuration_valid(self):
        self.backup_opt.lvm_srcvol = ''
        self.backup_opt.lvm_volgroup = ''
        self.assertFalse(lvm.validate_lvm_params(self.backup_opt))

    def test_return_false_on_lvm_not_required(self):
        self.assertTrue(lvm.validate_lvm_params(self.backup_opt))

    def test_raises_Exception_on_snapperm_invalid(self):
        self.backup_opt.lvm_snapperm = 'squeezeme'
        self.assertRaises(Exception, lvm.validate_lvm_params,
                          self.backup_opt)  # noqa

    def test_raises_Exception_on_no_pathtobackup(self):
        self.backup_opt.path_to_backup = ''
        self.assertRaises(Exception, lvm.validate_lvm_params,
                          self.backup_opt)  # noqa

    def test_raises_Exception_on_no_lvmsrcvol(self):
        self.backup_opt.lvm_srcvol = ''
        self.assertRaises(Exception, lvm.validate_lvm_params,
                          self.backup_opt)  # noqa

    def test_raises_Exception_on_no_lvmvolgrp(self):
        self.backup_opt.lvm_volgroup = ''
        self.assertRaises(Exception, lvm.validate_lvm_params,
                          self.backup_opt)  # noqa


class Test_umount(unittest.TestCase):
    @patch('freezer.snapshot.lvm.subprocess.Popen')
    @patch('freezer.snapshot.lvm.os')
    def test_return_none_on_success(self, mock_os, mock_popen):
        mock_process = mock.Mock()
        mock_process.communicate.return_value = '', ''
        mock_process.returncode = 0
        mock_popen.return_value = mock_process
        mock_os.rmdir.return_value = None
        self.assertIsNone(lvm._umount('path'))

    @patch('freezer.snapshot.lvm.subprocess.Popen')
    @patch('freezer.snapshot.lvm.os')
    def test_raises_on_popen_returncode_not_0(self, mock_os, mock_popen):
        mock_process = mock.Mock()
        mock_process.communicate.return_value = '', ''
        mock_process.returncode = 1
        mock_popen.return_value = mock_process
        mock_os.rmdir.return_value = None
        self.assertRaises(Exception, lvm._umount, 'path')  # noqa


class Test_lvremove(unittest.TestCase):
    @patch('freezer.snapshot.lvm.subprocess.Popen')
    def test_return_none_on_success(self, mock_popen):
        mock_process = mock.Mock()
        mock_process.communicate.return_value = '', ''
        mock_process.returncode = 0
        mock_popen.return_value = mock_process
        self.assertIsNone(lvm._lvremove('logicalvolume'))

    @patch('freezer.snapshot.lvm.subprocess.Popen')
    def test_raises_on_popen_returncode_not_0(self, mock_popen):
        mock_process = mock.Mock()
        mock_process.communicate.return_value = '', ''
        mock_process.returncode = 1
        mock_popen.return_value = mock_process
        self.assertRaises(Exception, lvm._lvremove, 'path')  # noqa



# (c) Copyright 2016 Hewlett-Packard Development Enterprise, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from concurrent import futures
import os

from oslo_config import cfg
from oslo_log import log
from oslo_serialization import jsonutils as json

from freezer.common import client_manager
from freezer.engine import engine
from freezer.engine.tar import tar
from freezer.utils import utils

import tempfile

LOG = log.getLogger(__name__)
CONF = cfg.CONF


class NovaEngine(engine.BackupEngine):

    def __init__(self, storage, **kwargs):
        super(NovaEngine, self).__init__(storage=storage)
        self.client = client_manager.get_client_manager(CONF)
        self.nova = self.client.create_nova()
        self.glance = self.client.create_glance()
        self.cinder = self.client.create_cinder()
        self.neutron = self.client.create_neutron()
        self.encrypt_pass_file = kwargs.get('encrypt_key')
        self.exclude = kwargs.get('exclude')
        self.server_info = None
        self.openssl_path = None
        self.compression_algo = 'gzip'
        self.is_windows = None
        self.dry_run = kwargs.get('dry_run', False)
        self.max_segment_size = kwargs.get('max_segment_size')
        self.storage = storage
        self.dereference_symlink = kwargs.get('symlinks')

    @property
    def name(self):
        return "nova"

    def stream_image(self, pipe):
        """Reading bytes from a pipe and converting it to a stream-like"""
        try:
            while True:
                chunk = pipe.recv_bytes()
                yield chunk
        except EOFError:
            pass

    def get_nova_tenant(self, project_id):
        # Load info about tenant instances
        if self.storage._type == 'swift':
            swift_connection = self.client.create_swift()
            headers, data = swift_connection.get_object(
                self.storage.storage_path,
                "project_" + project_id)
        elif self.storage._type == 's3':
            bucket_name, object_name = self.get_storage_info(project_id)
            data = self.storage.get_object(
                bucket_name=bucket_name,
                key=object_name
            )['Body'].read()
        elif self.storage._type in ['local', 'ssh', 'ftp', 'ftps']:
            backup_basepath = os.path.join(self.storage.storage_path,
                                           'project_' + project_id)
            with self.storage.open(backup_basepath, 'rb') as backup_file:
                data = backup_file.readline()
        elif self.storage._type in ['ftp', 'ftps']:
            backup_basepath = os.path.join(self.storage.storage_path,
                                           'project_' + project_id)
            file = tempfile.NamedTemporaryFile('wb', delete=True)
            self.storage.get_file(backup_basepath, file.name)
            with open(file.name) as f:
                data = f.readline()
            LOG.info("get_nova_tenant download {0}".format(data))

        return json.loads(data)

    def restore_nova_tenant(self, project_id, hostname_backup_name,
                            overwrite, recent_to_date):

        instance_ids = self.get_nova_tenant(project_id)
        for instance_id in instance_ids:
            LOG.info("Restore nova instance ID: {0} from container {1}".
                     format(instance_id, self.storage.storage_path))
            backup_name = os.path.join(hostname_backup_name,
                                       instance_id)
            self.restore(
                hostname_backup_name=backup_name,
                restore_resource=instance_id,
                overwrite=overwrite,
                recent_to_date=recent_to_date)

    def restore_level(self, restore_resource, read_pipe, backup, except_queue):
        try:
            metadata = backup.metadata()
            if (not self.encrypt_pass_file and
                    metadata.get("encryption", False)):
                raise Exception("Cannot restore encrypted backup without key")
            engine_metadata = backup.engine_metadata()
            server_info = metadata.get('server', {})
            length = int(engine_metadata.get('length'))
            available_networks = server_info.get('addresses')
            nova_networks = self.neutron.list_networks()['networks']

            net_names = [network for network, _ in
                         available_networks.items()]
            match_networks = [{"net-id": network.get('id')} for network in
                              nova_networks
                              if network.get('name') in net_names]

            stream = self.stream_image(read_pipe)
            data = utils.ReSizeStream(stream, length, 1)
            image = self.client.create_image(
                "Restore: {0}".format(
                    server_info.get('name', server_info.get('id', None))
                ),
                'bare',
                'raw',
                data=data
            )

            if self.encrypt_pass_file:
                try:
                    tmpdir = tempfile.mkdtemp()
                except Exception:
                    LOG.error("Unable to create a tmp directory")
                    raise

                tar_engine = tar.TarEngine(self.compression_algo,
                                           self.dereference_symlink,
                                           self.exclude, self.storage,
                                           self.max_segment_size,
                                           self.encrypt_pass_file,
                                           self.dry_run)

                tar_engine.restore_level(tmpdir, read_pipe, backup,
                                         except_queue)

            utils.wait_for(
                NovaEngine.image_active,
                1,
                CONF.timeout,
                message="Waiting for image to finish uploading {0} and become"
                        " active".format(image.id),
                kwargs={"glance_client": self.glance, "image_id": image.id}
            )
            server = self.nova.servers.create(
                name=server_info.get('name'),
                flavor=server_info['flavor']['id'],
                image=image.id,
                nics=match_networks
            )
            return server
        except Exception as e:
            LOG.exception(e)
            except_queue.put(e)
            raise

    def backup_nova_tenant(self, project_id, hostname_backup_name,
                           no_incremental, max_level, always_level,
                           restart_always_level):
        instance_ids = [server.id for server in
                        self.nova.servers.list(detailed=False)]
        data = json.dumps(instance_ids)
        LOG.info("Saving information about instances {0}".format(data))

        if self.storage._type == 'swift':
            swift_connection = self.client.create_swift()
            swift_connection.put_object(self.storage.storage_path,
                                        "project_{0}".format(project_id),
                                        data)
        elif self.storage._type == 's3':
            bucket_name, object_name = self.get_storage_info(project_id)
            self.storage.put_object(
                bucket_name=bucket_name,
                key=object_name,
                data=data
            )
        elif self.storage._type in ['local', 'ssh']:
            backup_basepath = os.path.join(self.storage.storage_path,
                                           "project_" + project_id)
            with self.storage.open(backup_basepath, 'wb') as backup_file:
                backup_file.write(data)
        elif self.storage._type in ['ftp', 'ftps']:
            backup_basepath = os.path.join(self.storage.storage_path,
                                           'project_' + project_id)
            file = tempfile.NamedTemporaryFile('wb', delete=True)
            with open(file.name, 'wb') as f:
                f.write(data)
            LOG.info("backup_nova_tenant data={0}".format(data))
            self.storage.put_file(file.name, backup_basepath)

        executor = futures.ThreadPoolExecutor(
            max_workers=len(instance_ids))
        futures_list = []
        for instance_id in instance_ids:
            LOG.info("Backup nova instance ID: {0} to container {1}".
                     format(instance_id, self.storage.storage_path))
            backup_name = os.path.join(hostname_backup_name,
                                       instance_id)

            futures_list.append(executor.submit(
                self.backup,
                backup_resource=instance_id,
                hostname_backup_name=backup_name,
                no_incremental=no_incremental,
                max_level=max_level,
                always_level=always_level,
                restart_always_level=restart_always_level))

        futures.wait(futures_list, CONF.timeout)

    def get_storage_info(self, project_id):
        if self.storage.get_object_prefix() != '':
            object_name = "{0}/project_{1}".format(
                self.storage.get_object_prefix(),
                project_id
            )
        else:
            object_name = "project_{0}".format(project_id)
        return self.storage.get_bucket_name(), object_name

    def backup_data(self, backup_resource, manifest_path):
        server = self.nova.servers.get(backup_resource)
        if not server:
            raise Exception("Server not found {0}".format(backup_resource))

        def instance_finish_task():
            server = self.nova.servers.get(backup_resource)
            return not server.__dict__['OS-EXT-STS:task_state']

        utils.wait_for(
            instance_finish_task, 1, CONF.timeout,
            message="Waiting for instance {0} to finish {1} to start the "
                    "snapshot process".format(
                        backup_resource,
                        server.__dict__['OS-EXT-STS:task_state']
                    )
        )
        image_id = self.nova.servers.create_image(
            server,
            "snapshot_of_{0}".format(backup_resource)
        )
        image = self.glance.images.get(image_id)
        if not image:
            raise Exception(
                "Image {0} is not created or can't be found.".format(image_id)
            )
        # wait a bit for the snapshot to be taken and completely uploaded
        # to glance.
        utils.wait_for(
            NovaEngine.image_active,
            1,
            100,
            message="Waiting for instance {0} snapshot to become "
                    "active".format(backup_resource),
            kwargs={"glance_client": self.glance, "image_id": image_id}
        )

        image = self.glance.images.get(image_id)
        image_temporary_snapshot_id = None
        copied_volume = None
        image_info = getattr(server, "image", None)
        if image_info is not None and isinstance(image_info, dict):
            LOG.info('Image type instance backup')
            boot_device_type = "image"
            stream = self.client.download_image(image)
        else:
            LOG.info('Volume or snapshot type instance backup')
            boot_device_type = "volume"
            image_block_mapping_info = image.get("block_device_mapping")
            image_block_mapping = json.loads(image_block_mapping_info) \
                if image_block_mapping_info else None
            image_temporary_snapshot_id = \
                image_block_mapping[0]['snapshot_id'] \
                if image_block_mapping else None
            copied_volume = self.client.do_copy_volume(
                self.cinder.volume_snapshots.get(
                    image_temporary_snapshot_id))
            LOG.debug("Deleting temporary glance image "
                      "generated by snapshot")
            self.glance.images.delete(image.id)
            LOG.debug("Creation temporary glance image")
            image = self.client.make_glance_image(
                copied_volume.id, copied_volume)
            LOG.debug("Download temporary glance image {0}".format(image.id))
            stream = self.client.download_image(image)

        LOG.info("Uploading image to storage path")
        headers = {"server_name": server.name,
                   "flavour_id": str(server.flavor.get('id')),
                   'length': str(len(stream)),
                   "boot_device_type": boot_device_type}
        self.set_tenant_meta(manifest_path, headers)
        for chunk in stream:
            yield chunk

        if self.encrypt_pass_file:
            tar_engine = tar.TarEngine(self.compression_algo,
                                       self.dereference_symlink,
                                       self.exclude, self.storage,
                                       self.max_segment_size,
                                       self.encrypt_pass_file, self.dry_run)

            for data_chunk in tar_engine.backup_data('.', manifest_path):
                yield data_chunk

        if image_temporary_snapshot_id is not None:
            LOG.info("Deleting temporary snapshot {0}"
                     .format(image_temporary_snapshot_id))
            self.cinder.volume_snapshots.delete(image_temporary_snapshot_id)
        if copied_volume is not None:
            LOG.info("Deleting temporary copied volume {0}"
                     .format(copied_volume.id))
            self.cinder.volumes.delete(copied_volume)

        LOG.info("Deleting temporary image {0}".format(image.id))
        self.glance.images.delete(image.id)

    @staticmethod
    def image_active(glance_client, image_id):
        """Check if the image is in the active state or not"""
        image = glance_client.images.get(image_id)
        return image.status == 'active'

    def metadata(self, backup_resource):
        """Construct metadata"""
        server_info = self.nova.servers.get(backup_resource).to_dict()

        return {
            "engine_name": self.name,
            "server": server_info,
            "encryption": bool(self.encrypt_pass_file)
        }

    def set_tenant_meta(self, path, metadata):
        """push data to the manifest file"""
        with open(path, 'wb') as fb:
            fb.writelines(json.dumps(metadata))

    def get_tenant_meta(self, path):
        with open(path, 'rb') as fb:
            json.loads(fb.read())



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from cinderclient import exceptions
from os_brick.initiator import connector
from oslo_concurrency import processutils

from freezer.engine.osbrick import brick_utils
from freezer.engine.osbrick import volume_actions as actions


class Client(object):
    version = '1.1.0'

    def __init__(self, volumes_client=None):
        self.volumes_client = volumes_client

    def _brick_get_connector(self, protocol, driver=None,
                             execute=processutils.execute,
                             use_multipath=False,
                             device_scan_attempts=3,
                             *args, **kwargs):
        """Wrapper to get a brick connector object.

        This automatically populates the required protocol as well
        as the root_helper needed to execute commands.
        """
        return connector.InitiatorConnector.factory(
            protocol,
            brick_utils.get_root_helper(),
            driver=driver,
            execute=execute,
            use_multipath=use_multipath,
            device_scan_attempts=device_scan_attempts,
            *args, **kwargs)

    def get_connector(self, multipath=False, enforce_multipath=False):
        conn_prop = connector.get_connector_properties(
            brick_utils.get_root_helper(),
            brick_utils.get_my_ip(),
            multipath=multipath,
            enforce_multipath=(enforce_multipath),
            execute=processutils.execute)
        return conn_prop

    def attach(self, volume_id, hostname, mountpoint=None, mode='rw',
               multipath=False, enforce_multipath=False):

        # Check protocol type of storage backend.
        with actions.VerifyProtocol(self.volumes_client, volume_id) as cmd:
            # Retrieve vol-host attribute of volume.
            volume_info = self.volumes_client.volumes.get(volume_id)
            volume_capabilities = self.volumes_client.capabilities.get(
                volume_info.__dict__['os-vol-host-attr:host'])
            # Retrieve storage_protocol from storage backend capabilities.
            protocol = volume_capabilities.storage_protocol.upper()
            cmd.verify(protocol)

        # Reserve volume before attachment
        with actions.Reserve(self.volumes_client, volume_id) as cmd:
            cmd.reserve()

        with actions.InitializeConnection(
                self.volumes_client, volume_id) as cmd:
            connection = cmd.initialize(self, multipath, enforce_multipath)

        with actions.ConnectVolume(self.volumes_client, volume_id) as cmd:
            brick_connector = self._brick_get_connector(
                protocol, do_local_attach=True)
            device_info = cmd.connect(brick_connector,
                                      connection['data'],
                                      mountpoint, mode, hostname)
            return device_info

    def detach(self, volume_id, attachment_uuid=None, multipath=False,
               enforce_multipath=False, device_info=None):

        with actions.BeginDetach(self.volumes_client, volume_id) as cmd:
            cmd.reserve()

        with actions.InitializeConnectionForDetach(
                self.volumes_client, volume_id) as cmd:
            connection = cmd.initialize(self, multipath, enforce_multipath)

        brick_connector = self._brick_get_connector(
            connection['driver_volume_type'], do_local_attach=True)

        with actions.DisconnectVolume(self.volumes_client, volume_id) as cmd:
            cmd.disconnect(brick_connector, connection['data'], device_info)

        with actions.DetachVolume(self.volumes_client, volume_id) as cmd:
            cmd.detach(self, attachment_uuid, multipath, enforce_multipath)

    def get_volume_paths(self, volume_id, use_multipath=False):
        """Gets volume paths on the system for a specific volume."""
        conn_props = self.get_connector(multipath=use_multipath)
        vols = self.volumes_client.volumes.list()
        vol_in_use = False
        vol_found = False
        for vol in vols:
            if (volume_id == vol.id or volume_id == vol.name):
                vol_found = True
                if vol.status == "in-use":
                    vol_in_use = True
                    # Make sure the volume ID is used and not the name
                    volume_id = vol.id
                break

        if not vol_found:
            msg = "No volume with a name or ID of '%s' exists." % volume_id
            raise exceptions.CommandError(msg)

        paths = []
        if vol_in_use:
            conn_info = self.volumes_client.volumes.initialize_connection(
                volume_id, conn_props)
            protocol = conn_info['driver_volume_type']
            conn = self._brick_get_connector(protocol,
                                             use_multipath=use_multipath)
            paths = conn.get_volume_paths(conn_info['data'])

        return paths

    def get_all_volume_paths(self, protocol, use_multipath=False):
        """Gets all volume paths on the system for a given protocol."""
        conn = self._brick_get_connector(protocol, use_multipath=use_multipath)
        paths = conn.get_all_available_volumes()

        return paths
